import datetime as dt
import os
import warnings
from datetime import datetime

import pyarrow as pa
from dagster import (
    AssetExecutionContext,
    AssetIn,
    DailyPartitionsDefinition,
    DimensionPartitionMapping,
    MultiPartitionMapping,
    MultiPartitionsDefinition,
    MultiToSingleDimensionPartitionMapping,
    SpecificPartitionsPartitionMapping,
    StaticPartitionMapping,
    StaticPartitionsDefinition,
    asset,
    materialize,
)
from deltalake import DeltaTable

from dagster_delta import DeltaLakePyarrowIOManager

warnings.filterwarnings("ignore")

daily_partitions_def = DailyPartitionsDefinition(
    start_date="2022-01-01",
    end_date="2022-01-10",
)

letter_partitions_def = StaticPartitionsDefinition(["a", "b", "c"])

color_partitions_def = StaticPartitionsDefinition(["red", "blue", "yellow"])

multi_partition_with_letter = MultiPartitionsDefinition(
    partitions_defs={
        "date": daily_partitions_def,
        "letter": letter_partitions_def,
    },
)

multi_partition_with_color = MultiPartitionsDefinition(
    partitions_defs={
        "date": daily_partitions_def,
        "color": color_partitions_def,
    },
)


@asset(
    key_prefix=["my_schema"],
)
def asset_1() -> pa.Table:
    return pa.Table.from_pydict(
        {
            "value": [1],
            "b": [1],
        },
    )


@asset(
    key_prefix=["my_schema"],
)
def asset_2(asset_1: pa.Table) -> pa.Table:
    return asset_1


# case: we have multiple partitions
@asset(
    key_prefix=["my_schema"],
    partitions_def=multi_partition_with_color,
    metadata={
        "partition_expr": {
            "date": "date_column",
            "color": "color_column",
        },
    },
)
def multi_partitioned_asset_1(context: AssetExecutionContext) -> pa.Table:
    color, date = context.partition_key.split("|")
    date_parsed = dt.datetime.strptime(date, "%Y-%m-%d").date()

    return pa.Table.from_pydict(
        {
            "date_column": [date_parsed],
            "value": [1],
            "b": [1],
            "color_column": [color],
        },
    )


# Multi-to-multi asset is supported
@asset(
    key_prefix=["my_schema"],
    partitions_def=multi_partition_with_color,
    metadata={
        "partition_expr": {
            "date": "date_column",
            "color": "color_column",
        },
    },
)
def multi_partitioned_asset_2(multi_partitioned_asset_1: pa.Table) -> pa.Table:
    return multi_partitioned_asset_1


@asset(
    key_prefix=["my_schema"],
)
def non_partitioned_asset(multi_partitioned_asset_1: pa.Table) -> pa.Table:
    return multi_partitioned_asset_1


# Multi-to-single asset is supported through MultiToSingleDimensionPartitionMapping
@asset(
    key_prefix=["my_schema"],
    partitions_def=daily_partitions_def,
    ins={
        "multi_partitioned_asset": AssetIn(
            ["my_schema", "multi_partitioned_asset_1"],
            partition_mapping=MultiToSingleDimensionPartitionMapping(
                partition_dimension_name="date",
            ),
        ),
    },
    metadata={
        "partition_expr": "date_column",
    },
)
def single_partitioned_asset_date(multi_partitioned_asset: pa.Table) -> pa.Table:
    return multi_partitioned_asset


@asset(
    key_prefix=["my_schema"],
    partitions_def=color_partitions_def,
    ins={
        "multi_partitioned_asset": AssetIn(
            ["my_schema", "multi_partitioned_asset_1"],
            partition_mapping=MultiToSingleDimensionPartitionMapping(
                partition_dimension_name="color",
            ),
        ),
    },
    metadata={
        "partition_expr": "color_column",
    },
)
def single_partitioned_asset_color(multi_partitioned_asset: pa.Table) -> pa.Table:
    return multi_partitioned_asset


@asset(
    partitions_def=multi_partition_with_letter,
    key_prefix=["my_schema"],
    metadata={"partition_expr": {"date": "date_column", "letter": "letter"}},
    ins={
        "multi_partitioned_asset": AssetIn(
            ["my_schema", "multi_partitioned_asset_1"],
            partition_mapping=MultiPartitionMapping(
                {
                    "color": DimensionPartitionMapping(
                        dimension_name="letter",
                        partition_mapping=StaticPartitionMapping(
                            {"blue": "a", "red": "b", "yellow": "c"},
                        ),
                    ),
                    "date": DimensionPartitionMapping(
                        dimension_name="date",
                        partition_mapping=SpecificPartitionsPartitionMapping(
                            ["2022-01-01", "2024-01-01"],
                        ),
                    ),
                },
            ),
        ),
    },
)
def mapped_multi_partition(
    context: AssetExecutionContext,
    multi_partitioned_asset: pa.Table,
) -> pa.Table:
    _, letter = context.partition_key.split("|")

    table_ = multi_partitioned_asset.append_column("letter", pa.array([letter]))
    return table_


def test_unpartitioned_asset_to_unpartitioned_asset(
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    res = materialize([asset_1, asset_2], resources=resource_defs)
    assert res.success

    asset_1_data = res.asset_value(asset_1.key)
    asset_2_data = res.asset_value(asset_2.key)
    assert asset_1_data == asset_2_data


def test_multi_partitioned_to_multi_partitioned_asset(
    tmp_path,
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    multi_partitioned_asset_1_data_all = []
    multi_partitioned_asset_2_data_all = []

    for partition_key in ["red|2022-01-01", "red|2022-01-02", "red|2022-01-03"]:
        res = materialize(
            [multi_partitioned_asset_1, multi_partitioned_asset_2],
            partition_key=partition_key,
            resources=resource_defs,
        )
        assert res.success

        multi_partitioned_asset_1_data = res.asset_value(multi_partitioned_asset_1.key)
        multi_partitioned_asset_2_data = res.asset_value(multi_partitioned_asset_2.key)
        assert multi_partitioned_asset_1_data.select(
            multi_partitioned_asset_1_data.column_names,
        ) == multi_partitioned_asset_2_data.select(
            multi_partitioned_asset_1_data.column_names,
        )

        multi_partitioned_asset_1_data_all.append(multi_partitioned_asset_1_data)
        multi_partitioned_asset_2_data_all.append(multi_partitioned_asset_2_data)

    dt = DeltaTable(os.path.join(str(tmp_path), "/".join(multi_partitioned_asset_1.key.path)))

    assert dt.to_pyarrow_table().sort_by("date_column") == pa.concat_tables(
        multi_partitioned_asset_1_data_all,
    )

    dt = DeltaTable(os.path.join(str(tmp_path), "/".join(multi_partitioned_asset_2.key.path)))
    assert dt.metadata().partition_columns == ["color_column", "date_column"]
    assert dt.to_pyarrow_table().sort_by("date_column") == pa.concat_tables(
        multi_partitioned_asset_2_data_all,
    )


def test_multi_partitioned_to_single_partitioned_asset_colors(
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    multi_partitioned_asset_1_data_all = []

    for partition_key in ["red|2022-01-01", "blue|2022-01-01", "yellow|2022-01-01"]:
        res = materialize(
            [multi_partitioned_asset_1],
            partition_key=partition_key,
            resources=resource_defs,
        )
        assert res.success

        multi_partitioned_asset_1_data = res.asset_value(multi_partitioned_asset_1.key)
        multi_partitioned_asset_1_data_all.append(multi_partitioned_asset_1_data)


def test_multi_partitioned_to_single_partitioned_asset_dates(
    tmp_path,
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    multi_partitioned_asset_1_data_all = []

    for partition_key in ["red|2022-01-01", "red|2022-01-02", "red|2022-01-03"]:
        res = materialize(
            [multi_partitioned_asset_1],
            partition_key=partition_key,
            resources=resource_defs,
        )
        assert res.success

        multi_partitioned_asset_1_data = res.asset_value(multi_partitioned_asset_1.key)
        multi_partitioned_asset_1_data_all.append(multi_partitioned_asset_1_data)

    res = materialize(
        [multi_partitioned_asset_1, single_partitioned_asset_color],
        partition_key="red",
        resources=resource_defs,
        selection=[single_partitioned_asset_color],
    )
    assert res.success

    single_partitioned_asset_color_data = res.asset_value(single_partitioned_asset_color.key)
    cols = single_partitioned_asset_color_data.column_names
    assert single_partitioned_asset_color_data.select(cols).sort_by(
        "date_column",
    ) == pa.concat_tables(
        multi_partitioned_asset_1_data_all,
    ).select(cols).sort_by("date_column")

    dt = DeltaTable(os.path.join(str(tmp_path), "/".join(single_partitioned_asset_color.key.path)))
    assert dt.metadata().partition_columns == ["color_column"]

    assert single_partitioned_asset_color_data == dt.to_pyarrow_table()


def test_multi_partitioned_to_non_partitioned_asset(
    tmp_path,
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    multi_partitioned_asset_1_data_all = []

    for partition_key in ["red|2022-01-01", "red|2022-01-02", "red|2022-01-03"]:
        res = materialize(
            [multi_partitioned_asset_1],
            partition_key=partition_key,
            resources=resource_defs,
        )
        assert res.success

        multi_partitioned_asset_1_data = res.asset_value(multi_partitioned_asset_1.key)
        multi_partitioned_asset_1_data_all.append(multi_partitioned_asset_1_data)

    res = materialize(
        [multi_partitioned_asset_1, non_partitioned_asset],
        resources=resource_defs,
        selection=[non_partitioned_asset],
    )
    assert res.success

    non_partitioned_asset_data = res.asset_value(non_partitioned_asset.key)
    cols = non_partitioned_asset_data.column_names

    assert non_partitioned_asset_data.select(cols).sort_by("date_column") == pa.concat_tables(
        multi_partitioned_asset_1_data_all,
    ).select(cols).sort_by("date_column")

    dt = DeltaTable(os.path.join(str(tmp_path), "/".join(non_partitioned_asset.key.path)))
    assert dt.metadata().partition_columns == []

    assert non_partitioned_asset_data == dt.to_pyarrow_table()


def test_multi_partitioned_to_multi_partitioned_with_different_dimensions(
    tmp_path,
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    for partition_key in ["red|2022-01-01", "blue|2022-01-01", "yellow|2022-01-01"]:
        res = materialize(
            [multi_partitioned_asset_1],
            partition_key=partition_key,
            resources=resource_defs,
        )
        assert res.success

    res = materialize(
        [multi_partitioned_asset_1, mapped_multi_partition],
        partition_key="2022-01-01|a",
        resources=resource_defs,
        selection=[mapped_multi_partition],
    )

    date_parsed = datetime.strptime("2022-01-01", "%Y-%m-%d").date()

    expected = pa.Table.from_pydict(
        {
            "value": [1],
            "b": [1],
            "color_column": ["blue"],
            "date_column": [date_parsed],
            "letter": ["a"],
        },
    )

    dt = DeltaTable(os.path.join(str(tmp_path), "/".join(mapped_multi_partition.key.path)))
    assert dt.metadata().partition_columns == ["date_column", "letter"]
    assert expected == dt.to_pyarrow_table()



from typing import Optional, Union

from dagster import (
    InputContext,
    MultiPartitionsDefinition,
    OutputContext,
)

try:
    from dagster._core.definitions.partitions.definition.time_window import (
        TimeWindowPartitionsDefinition,
    )
except ModuleNotFoundError:
    from dagster._core.definitions.time_window_partitions import (
        TimeWindowPartitionsDefinition,
    )


def extract_date_format_from_partition_definition(
    context: Union[OutputContext, InputContext],
) -> Optional[dict[str, str]]:
    """Gets the date format from the partition definition if there is a TimeWindowPartitionsDefinition present (nested or not), to be used to properly compare with columns
    in the delta table which are not a datetime object. Returns None if no TimeWindowPartitionsDefinition were present.
    """
    if isinstance(context, InputContext):
        if context.has_asset_partitions:
            if context.upstream_output is not None:
                partition_expr = context.upstream_output.definition_metadata["partition_expr"]  # type: ignore
                partitions_definition = context.asset_partitions_def
            else:
                raise ValueError(
                    "'partition_expr' should have been set in the metadata of the incoming asset since it has a partition definition.",
                )
        else:
            return None
    elif isinstance(context, OutputContext):
        if context.has_asset_partitions:
            if (
                context.definition_metadata is not None
                and "partition_expr" in context.definition_metadata
            ):
                partition_expr = context.definition_metadata["partition_expr"]
            else:
                raise ValueError(
                    "'partition_expr' should have been set in the metadata of the incoming asset since it has a partition definition.",
                )
            partitions_definition = context.asset_partitions_def
        else:
            return None
    if partition_expr is None or partitions_definition is None:
        return None

    date_format: dict[str, str] = {}
    if isinstance(partitions_definition, TimeWindowPartitionsDefinition):
        if isinstance(partition_expr, str):
            date_format[partition_expr] = partitions_definition.fmt  # type: ignore
        else:
            raise ValueError(
                "Single partition definition provided, so partion_expr needs to be a string",
            )
    elif isinstance(partitions_definition, MultiPartitionsDefinition):
        if isinstance(partition_expr, dict):
            for partition_dims_definition in partitions_definition.partitions_defs:
                if isinstance(
                    partition_dims_definition.partitions_def,
                    TimeWindowPartitionsDefinition,
                ):
                    partition_expr_name = partition_expr.get(partition_dims_definition.name)
                    if partition_expr_name is None:
                        raise ValueError(
                            f"Partition_expr mapping is invalid. Partition_dimension :{partition_dims_definition.name} not found in partition_expr: {partition_expr}.",
                        )
                    date_format[partition_expr_name] = partition_dims_definition.partitions_def.fmt  # type: ignore[attr-defined]
        else:
            raise ValueError(
                "MultiPartitionsDefinition provided, so partion_expr needs to be a dictionary mapping of {dimension: column}",
            )

    return date_format if len(date_format) else None



import os
from datetime import date, datetime

import pyarrow as pa
import pyarrow.compute as pc
import pytest
from dagster import (
    Out,
    TimeWindow,
    graph,
    op,
)
from dagster._core.storage.db_io_manager import TablePartitionDimension
from deltalake import DeltaTable
from deltalake.schema import Field, PrimitiveType, Schema

from dagster_delta import DeltaLakePyarrowIOManager, LocalConfig, WriteMode
from dagster_delta._handler.utils import partition_dimensions_to_dnf

TablePartitionDimension(
    partitions=TimeWindow(datetime(2020, 1, 2), datetime(2020, 2, 3)),
    partition_expr="my_timestamp_col",
)


@pytest.fixture
def test_schema() -> Schema:
    fields = [
        Field(name="string_col", type=PrimitiveType("string")),
        Field(name="date_col", type=PrimitiveType("date")),
        Field(name="timestamp_col", type=PrimitiveType("timestamp")),
    ]
    return Schema(fields=fields)


def test_partition_dimensions_to_dnf(test_schema) -> None:
    parts = [
        TablePartitionDimension(
            partitions=TimeWindow(datetime(2020, 1, 2), datetime(2020, 2, 3)),
            partition_expr="timestamp_col",
        ),
    ]
    dnf = partition_dimensions_to_dnf(parts, test_schema)
    assert dnf == [("timestamp_col", "=", datetime(2020, 1, 2, 0, 0, 0))]

    parts = [
        TablePartitionDimension(
            partitions=TimeWindow(datetime(2020, 1, 2), datetime(2020, 2, 3)),
            partition_expr="date_col",
        ),
    ]
    dnf = partition_dimensions_to_dnf(parts, test_schema)
    assert dnf == [("date_col", "=", date(2020, 1, 2))]


@op(out=Out(metadata={"schema": "a_df"}))
def a_pa_table() -> pa.Table:
    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one"}))
def add_one(table: pa.Table) -> pa.Table:
    # Add one to each column in the table
    updated_columns = [pc.add_checked(table[column], pa.scalar(1)) for column in table.column_names]  # type: ignore
    updated_table = pa.Table.from_arrays(updated_columns, names=table.column_names)
    return updated_table


@graph
def add_one_to_dataset():
    add_one(a_pa_table())


@pytest.fixture
def io_manager_with_parquet_read_options(tmp_path) -> DeltaLakePyarrowIOManager:
    return DeltaLakePyarrowIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        mode=WriteMode.overwrite,
        parquet_read_options={"coerce_int96_timestamp_unit": "us"},
    )


def test_deltalake_io_manager_with_parquet_read_options(
    tmp_path,
    io_manager_with_parquet_read_options,
):
    resource_defs = {"io_manager": io_manager_with_parquet_read_options}

    job = add_one_to_dataset.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [2, 3, 4]



import os

import polars as pl
import pytest
from dagster import (
    Out,
    graph,
    op,
)
from deltalake import DeltaTable

from dagster_delta import DeltaLakePolarsIOManager, LocalConfig, WriteMode


@pytest.fixture
def io_manager(tmp_path) -> DeltaLakePolarsIOManager:
    return DeltaLakePolarsIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        mode=WriteMode.overwrite,
    )


@pytest.fixture
def io_manager_append(tmp_path) -> DeltaLakePolarsIOManager:
    return DeltaLakePolarsIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        mode=WriteMode.append,
    )


@pytest.fixture
def io_manager_ignore(tmp_path) -> DeltaLakePolarsIOManager:
    return DeltaLakePolarsIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        mode=WriteMode.ignore,
    )


@op(out=Out(metadata={"schema": "a_df"}))
def a_df() -> pl.DataFrame:
    return pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one"}))
def add_one(df: pl.DataFrame):  # noqa: ANN201
    return df + 1


@graph
def add_one_to_dataframe():
    add_one(a_df())


@graph
def just_a_df():
    a_df()


def test_deltalake_io_manager_with_ops_appended(tmp_path, io_manager_append):
    resource_defs = {"io_manager": io_manager_append}

    job = just_a_df.to_job(resource_defs=resource_defs)

    # run the job twice to ensure tables get appended
    expected_result1 = [1, 2, 3]

    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == expected_result1

        expected_result1.extend(expected_result1)


def test_deltalake_io_manager_with_ops_ignored(tmp_path, io_manager_ignore):
    resource_defs = {"io_manager": io_manager_ignore}

    job = just_a_df.to_job(resource_defs=resource_defs)

    # run the job 5 times to ensure tables gets ignored on each write
    for _ in range(5):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

    dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
    assert dt.version() == 0


@op(out=Out(metadata={"schema": "a_df", "mode": "append"}))
def a_df_custom() -> pl.DataFrame:
    return pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@graph
def add_one_to_dataframe_custom():
    add_one(a_df_custom())


def test_deltalake_io_manager_with_ops_mode_overridden(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    job = add_one_to_dataframe_custom.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted

    a_df_result = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    add_one_result = pl.DataFrame({"a": [2, 3, 4], "b": [5, 6, 7]})
    from polars.testing import assert_frame_equal

    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = pl.DataFrame(dt.to_pyarrow_table())
        assert_frame_equal(out_df, a_df_result, check_row_order=False)

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = pl.DataFrame(dt.to_pyarrow_table())
        assert_frame_equal(out_df, add_one_result, check_row_order=False)

        a_df_result = pl.concat([a_df_result, a_df_result])
        add_one_result = pl.concat([add_one_result, add_one_result])



from typing import Optional, Union

from dagster import ConfigurableResource
from deltalake import DeltaTable
from pydantic import Field

from dagster_delta.config import AzureConfig, ClientConfig, GcsConfig, LocalConfig, S3Config


class DeltaTableResource(ConfigurableResource):
    """Resource for interacting with a Delta table.

    Examples:
    ```python
        from dagster import Definitions, asset
        from dagster_delta import DeltaTableResource, LocalConfig

        @asset
        def my_table(delta_table: DeltaTableResource):
            df = delta_table.load().to_pandas()

        defs = Definitions(
            assets=[my_table],
            resources={
                "delta_table": DeltaTableResource(
                    url="/path/to/table",
                    storage_options=LocalConfig()
                )
            }
        )
    ```
    """

    url: str

    storage_options: Union[AzureConfig, S3Config, LocalConfig, GcsConfig] = Field(  # noqa: UP007
        discriminator="provider",
    )

    client_options: Optional[ClientConfig] = Field(
        default=None,
        description="Additional configuration passed to http client.",
    )

    version: Optional[int] = Field(default=None, description="Version to load delta table.")

    def load(self) -> DeltaTable:
        """Loads the table with passed configuration.

        Returns:
            DeltaTable: table
        """
        storage_options = self.storage_options.str_dict()
        client_options = self.client_options.str_dict() if self.client_options else {}
        options = {**storage_options, **client_options}

        table = DeltaTable(
            table_uri=self.url,
            storage_options=options,
            version=self.version,
        )
        return table



from dagster_delta._handler.utils.date_format import extract_date_format_from_partition_definition
from dagster_delta._handler.utils.dnf import partition_dimensions_to_dnf
from dagster_delta._handler.utils.predicates import create_predicate

__all__ = [
    "create_predicate",
    "extract_date_format_from_partition_definition",
    "partition_dimensions_to_dnf",
]



import logging
from abc import abstractmethod
from typing import Any, Generic, TypeVar, Union, cast

from arro3.core import RecordBatchReader, Table
from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
from dagster import (
    InputContext,
    MetadataValue,
    OutputContext,
    TableColumn,
    TableSchema,
)
from dagster._core.storage.db_io_manager import DbTypeHandler, TableSlice
from deltalake import CommitProperties, DeltaTable, QueryBuilder, WriterProperties, write_deltalake
from deltalake.exceptions import TableNotFoundError
from deltalake.schema import Schema
from deltalake.writer._conversion import _convert_arro3_schema_to_delta

from dagster_delta._handler.merge import merge_execute
from dagster_delta._handler.utils import (
    create_predicate,
    extract_date_format_from_partition_definition,
    partition_dimensions_to_dnf,
)
from dagster_delta.config import MergeConfig
from dagster_delta.io_manager.base import (
    TableConnection,
    _DeltaTableIOManagerResourceConfig,
)

T = TypeVar("T")

ArrowTypes = Union[RecordBatchReader, Table]
try:
    import pyarrow as pa

    ArrowTypes = Union[RecordBatchReader, Table, pa.Table, pa.RecordBatchReader]
except ImportError:
    pass


class DeltalakeBaseArrowTypeHandler(DbTypeHandler[T], Generic[T]):
    """Base TypeHandler implementation for arrow supported libraries used to handle deltalake IO."""

    @abstractmethod
    def from_arrow(
        self,
        obj: Union[ArrowStreamExportable, ArrowArrayExportable],
        target_type: type,
    ) -> T:
        """Abstract method to convert arrow to target type"""
        pass

    @abstractmethod
    def to_arrow(self, obj: T) -> RecordBatchReader:  # type: ignore
        """Abstract method to convert type to arrow"""
        pass

    @abstractmethod
    def get_output_stats(self, obj: T) -> dict[str, MetadataValue]:
        """Abstract method to return output stats"""
        pass

    @staticmethod
    def _find_keys_in_metadata(
        context: OutputContext,
        keys: list[str] = ["merge_predicate", "merge_operations_config"],
    ) -> dict[str, Any]:
        """Finds the keys in the metadata in the following order:

        It will find the merge_predicate or merge_operations_config in this order:
        1. Runtime metadata
        2. Definition metadata
        3. IO Manager config

        E.g., `merge_predicate` and `merge_operations_config`

        Args:
            context (OutputContext): The output context
            keys (list[str], optional): The keys to find in the metadata. Defaults to ["merge_predicate", "merge_operations_config"].

        Returns:
            dict[str, Any]: The metadata with the keys found
        """
        metadata_definition = context.definition_metadata or {}
        metadata_output = context.output_metadata or {}

        # Find each of the key in the definition or output metadata
        result = {}

        for key in keys:
            if key in metadata_output or {}:
                result[key] = metadata_output[key]
            else:
                result[key] = metadata_definition.get(key)

            # If it's a TextMetadataValue, cast it to string
            if isinstance(result[key], MetadataValue):
                result[key] = result[key].value

        return result

    def handle_output(
        self,
        context: OutputContext,
        table_slice: TableSlice,
        obj: T,
        connection: TableConnection,
    ):
        """Stores pyarrow types in Delta table."""
        logger = logging.getLogger()
        logger.setLevel("DEBUG")

        keys_from_metadata = self._find_keys_in_metadata(
            context,
            ["merge_predicate", "merge_operations_config"],
        )

        merge_predicate_from_metadata = keys_from_metadata.get("merge_predicate", None)
        merge_operations_config_from_metadata = keys_from_metadata.get(
            "merge_operations_config",
            None,
        )

        definition_metadata = context.definition_metadata or {}
        additional_table_config = definition_metadata.get("table_configuration", {})
        if connection.table_config is not None:
            table_config = additional_table_config | connection.table_config
        else:
            table_config = additional_table_config
        resource_config = context.resource_config or {}
        object_stats = self.get_output_stats(obj)

        data = self.to_arrow(obj=obj)
        delta_schema = Schema.from_arrow(_convert_arro3_schema_to_delta(data.schema))
        resource_config = cast(_DeltaTableIOManagerResourceConfig, context.resource_config)
        save_mode = definition_metadata.get("mode")
        main_save_mode = resource_config.get("mode")
        schema_mode = definition_metadata.get("schema_mode") or resource_config.get(
            "schema_mode",
        )
        if schema_mode is not None:
            schema_mode = str(schema_mode)

        writer_properties = resource_config.get("writer_properties")
        writer_properties = (
            WriterProperties(**writer_properties) if writer_properties is not None else None  # type: ignore
        )

        commit_properties = definition_metadata.get("commit_properties") or resource_config.get(
            "commit_properties",
        )
        commit_properties = (
            CommitProperties(**commit_properties) if commit_properties is not None else None  # type: ignore
        )
        merge_config = resource_config.get("merge_config")

        date_format = extract_date_format_from_partition_definition(context)

        if save_mode is not None:
            logger.debug(
                "IO manager mode overridden with the asset metadata mode, %s -> %s",
                main_save_mode,
                save_mode,
            )
            main_save_mode = save_mode
        logger.debug("Writing with mode: `%s`", main_save_mode)

        merge_stats = None
        partition_columns = None
        predicate = None

        if table_slice.partition_dimensions is not None:
            partition_filters = partition_dimensions_to_dnf(
                partition_dimensions=table_slice.partition_dimensions,
                table_schema=delta_schema,
                date_format=date_format,
            )
            if partition_filters is not None:
                ## Convert partition_filter to predicate
                predicate = create_predicate(partition_filters)

            partition_columns = [dim.partition_expr for dim in table_slice.partition_dimensions]

        if main_save_mode not in ["merge", "create_or_replace"]:
            if predicate is not None:
                logger.debug("Using explicit partition predicate: \n%s", predicate)
            write_deltalake(  # type: ignore
                table_or_uri=connection.table_uri,
                data=data,
                storage_options=connection.storage_options,
                mode=main_save_mode,  # type: ignore
                predicate=predicate,
                partition_by=partition_columns,
                schema_mode=schema_mode,  # type: ignore
                configuration=table_config,
                writer_properties=writer_properties,  # type: ignore
                commit_properties=commit_properties,
            )
        elif main_save_mode == "create_or_replace":
            DeltaTable.create(
                table_uri=connection.table_uri,
                schema=delta_schema,
                mode="overwrite",
                partition_by=partition_columns,
                configuration=table_config,
                storage_options=connection.storage_options,
            )
        else:
            if merge_config is None:
                raise ValueError(
                    "Merge Configuration should be provided when `mode = WriterMode.merge`",
                )
            try:
                dt = DeltaTable(connection.table_uri, storage_options=connection.storage_options)
            except TableNotFoundError:
                logger.debug("Creating a DeltaTable first before merging.")
                dt = DeltaTable.create(
                    table_uri=connection.table_uri,
                    schema=delta_schema,
                    partition_by=partition_columns,
                    configuration=table_config,
                    storage_options=connection.storage_options,
                    commit_properties=commit_properties,
                )
            merge_stats = merge_execute(
                dt,
                data,
                MergeConfig.model_validate(merge_config),
                writer_properties=writer_properties,
                commit_properties=commit_properties,
                merge_predicate_from_metadata=merge_predicate_from_metadata,
                merge_operations_config=merge_operations_config_from_metadata,
                partition_filters=partition_filters,
            )

        dt = DeltaTable(connection.table_uri, storage_options=connection.storage_options)

        output_metadata = {
            # "dagster/table_name": table_slice.table,
            "table_uri": MetadataValue.path(connection.table_uri),
            # "dagster/uri": MetadataValue.path(connection.table_uri),
            "dagster/column_schema": MetadataValue.table_schema(
                TableSchema(
                    columns=[
                        TableColumn(name=name, type=str(dtype))
                        for name, dtype in zip(
                            delta_schema.to_arrow().names,
                            delta_schema.to_arrow().types,
                        )
                    ],
                ),
            ),
            "table_version": MetadataValue.int(dt.version()),
            # **stats,
            **object_stats,
        }
        if merge_stats is not None:
            output_metadata["num_output_rows"] = MetadataValue.int(
                merge_stats.get("num_output_rows", 0),
            )
            output_metadata["merge_stats"] = MetadataValue.json(merge_stats)

        context.add_output_metadata(output_metadata)

    def load_input(
        self,
        context: InputContext,
        table_slice: TableSlice,
        connection: TableConnection,
    ) -> T:
        """Loads the input as a arro3 Table or RecordBatchReader."""
        table = DeltaTable(
            table_uri=connection.table_uri,
            storage_options=connection.storage_options,
        )
        logger = logging.getLogger()
        logger.setLevel("DEBUG")
        logger.debug("Connection timeout duration %s", connection.storage_options.get("timeout"))

        predicate = None
        if table_slice.partition_dimensions is not None:
            partition_filters = partition_dimensions_to_dnf(
                partition_dimensions=table_slice.partition_dimensions,
                table_schema=table.schema(),
                input_dnf=True,
            )

            if partition_filters is not None:
                ## Convert partition_filter to predicate
                predicate = create_predicate(partition_filters)

                logger.debug("Dataset input predicate %s", predicate)

        col_select = table_slice.columns if table_slice.columns is not None else "*"
        query = f"SELECT {col_select} FROM tbl"
        if predicate is not None:
            query = f"{query} WHERE {predicate}"
        data = QueryBuilder().register("tbl", table).execute(query)

        return self.from_arrow(data, context.dagster_type.typing_type)



import os

import pyarrow as pa
import pytest
from dagster import (
    Out,
    graph,
    op,
)
from deltalake import DeltaTable

from dagster_delta import DeltaLakePyarrowIOManager, LocalConfig


@pytest.fixture
def io_manager(tmp_path) -> DeltaLakePyarrowIOManager:
    return DeltaLakePyarrowIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
    )


@op(out=Out(metadata={"schema": "a_df"}))
def a_df() -> pa.Table:
    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one"}))
def add_one(df: pa.Table):  # noqa: ANN201
    return df.set_column(0, "a", pa.array([2, 3, 4]))


@graph
def add_one_to_dataframe():
    add_one(a_df())


def test_deltalake_io_manager_with_ops_rust_writer(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    job = add_one_to_dataframe.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [2, 3, 4]



import os
from datetime import datetime

import polars as pl
import pytest
from dagster import (
    AssetExecutionContext,
    AssetIn,
    DailyPartitionsDefinition,
    DynamicPartitionsDefinition,
    MultiPartitionKey,
    MultiPartitionsDefinition,
    Out,
    StaticPartitionsDefinition,
    asset,
    graph,
    instance_for_test,
    materialize,
    op,
)
from dagster._check import CheckError
from deltalake import DeltaTable, QueryBuilder

from dagster_delta import DeltaLakePolarsIOManager, LocalConfig, WriteMode
from dagster_delta.io_manager.base import DELTA_DATE_FORMAT


@pytest.fixture
def io_manager(tmp_path) -> DeltaLakePolarsIOManager:
    return DeltaLakePolarsIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        mode=WriteMode.overwrite,
    )


@op(out=Out(metadata={"schema": "a_df"}))
def a_df() -> pl.DataFrame:
    return pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one"}))
def add_one(df: pl.DataFrame) -> pl.DataFrame:
    return df + 1


@graph
def add_one_to_dataframe():
    add_one(a_df())


def test_deltalake_io_manager_with_ops(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    job = add_one_to_dataframe.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))

        out_df = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
        assert out_df["a"].to_pylist() == [2, 3, 4]


@asset(key_prefix=["my_schema"])
def b_df() -> pl.DataFrame:
    return pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@asset(key_prefix=["my_schema"])
def b_plus_one(b_df: pl.DataFrame) -> pl.DataFrame:
    return b_df + 1


@asset(key_prefix=["my_schema"])
def b_df_lazy() -> pl.LazyFrame:
    return pl.LazyFrame({"a": [1, 2, 3], "b": [4, 5, 6]})


@asset(key_prefix=["my_schema"])
def b_plus_one_lazy(b_df_lazy: pl.LazyFrame) -> pl.LazyFrame:
    print(b_df_lazy.collect_schema)
    return b_df_lazy.select(pl.all() + 1)


@pytest.mark.parametrize(
    ("asset1", "asset2", "asset1_path", "asset2_path"),
    [
        (b_df, b_plus_one, "b_df", "b_plus_one"),
        (b_df_lazy, b_plus_one_lazy, "b_df_lazy", "b_plus_one_lazy"),
    ],
)
def test_deltalake_io_manager_with_assets(
    tmp_path,
    io_manager,
    asset1,
    asset2,
    asset1_path,
    asset2_path,
):
    resource_defs = {"io_manager": io_manager}

    # materialize asset twice to ensure that tables get properly deleted
    for _ in range(2):
        res = materialize([asset1, asset2], resources=resource_defs)
        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "my_schema/" + asset1_path))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "my_schema/" + asset2_path))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [2, 3, 4]


def test_deltalake_io_manager_with_schema(tmp_path):
    @asset
    def my_df() -> pl.DataFrame:
        return pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

    @asset
    def my_df_plus_one(my_df: pl.DataFrame) -> pl.DataFrame:
        return my_df + 1

    io_manager = DeltaLakePolarsIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        schema="custom_schema",
    )

    resource_defs = {"io_manager": io_manager}

    # materialize asset twice to ensure that tables get properly deleted
    for _ in range(2):
        res = materialize([my_df, my_df_plus_one], resources=resource_defs)
        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "custom_schema/my_df"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "custom_schema/my_df_plus_one"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [2, 3, 4]


@asset(key_prefix=["my_schema"], ins={"b_df": AssetIn("b_df", metadata={"columns": ["a"]})})
def b_plus_one_columns(b_df: pl.DataFrame) -> pl.DataFrame:
    return b_df + 1


@asset(
    key_prefix=["my_schema"],
    ins={"b_df_lazy": AssetIn("b_df_lazy", metadata={"columns": ["a"]})},
)
def b_plus_one_columns_lazy(b_df_lazy: pl.LazyFrame) -> pl.LazyFrame:
    return b_df_lazy.select(pl.all() + 1)


@pytest.mark.parametrize(
    ("asset1", "asset2", "asset1_path", "asset2_path"),
    [
        (b_df, b_plus_one_columns, "b_df", "b_plus_one_columns"),
        (b_df_lazy, b_plus_one_columns_lazy, "b_df_lazy", "b_plus_one_columns_lazy"),
    ],
)
def test_loading_columns(tmp_path, io_manager, asset1, asset2, asset1_path, asset2_path):
    resource_defs = {"io_manager": io_manager}

    # materialize asset twice to ensure that tables get properly deleted
    for _ in range(2):
        res = materialize([asset1, asset2], resources=resource_defs)
        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "my_schema/" + asset1_path))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "my_schema/" + asset2_path))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == [2, 3, 4]

        assert out_df.shape[1] == 1


@op
def non_supported_type() -> int:
    return 1


@graph
def not_supported():
    non_supported_type()


def test_not_supported_type(io_manager):
    resource_defs = {"io_manager": io_manager}

    job = not_supported.to_job(resource_defs=resource_defs)

    with pytest.raises(
        CheckError,
        match="DeltaLakeIOManager does not have a handler for type '<class 'int'>'",
    ):
        job.execute_in_process()


@asset(
    partitions_def=DailyPartitionsDefinition(start_date="2022-01-01"),
    key_prefix=["my_schema"],
    metadata={"partition_expr": "time"},
    config_schema={"value": str},
)
def daily_partitioned(context: AssetExecutionContext) -> pl.DataFrame:
    partition = datetime.strptime(context.partition_key, DELTA_DATE_FORMAT).date()
    value = context.op_execution_context.op_config["value"]

    return pl.DataFrame(
        {
            "time": [partition, partition, partition],
            "a": [value, value, value],
            "b": [4, 5, 6],
        },
    )


@asset(
    partitions_def=DailyPartitionsDefinition(start_date="2022-01-01"),
    key_prefix=["my_schema"],
    metadata={"partition_expr": "time"},
    config_schema={"value": str},
)
def daily_partitioned_lazy(context: AssetExecutionContext) -> pl.LazyFrame:
    partition = datetime.strptime(
        context.partition_key,
        DELTA_DATE_FORMAT,
    ).date()
    value = context.op_execution_context.op_config["value"]

    return pl.LazyFrame(
        {
            "time": [partition, partition, partition],
            "a": [value, value, value],
            "b": [4, 5, 6],
        },
    )


@pytest.mark.parametrize(
    ("asset1", "asset1_path"),
    [
        (daily_partitioned, "daily_partitioned"),
        (daily_partitioned_lazy, "daily_partitioned_lazy"),
    ],
)
def test_time_window_partitioned_asset(tmp_path, io_manager, asset1, asset1_path):
    resource_defs = {"io_manager": io_manager}

    materialize(
        [asset1],
        partition_key="2022-01-01",
        resources=resource_defs,
        run_config={"ops": {"my_schema__" + asset1_path: {"config": {"value": "1"}}}},
    )

    dt = DeltaTable(os.path.join(tmp_path, "my_schema/" + asset1_path))
    out_df = dt.to_pyarrow_table()
    assert out_df["a"].to_pylist() == ["1", "1", "1"]

    materialize(
        [asset1],
        partition_key="2022-01-02",
        resources=resource_defs,
        run_config={"ops": {"my_schema__" + asset1_path: {"config": {"value": "2"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == ["1", "1", "1", "2", "2", "2"]

    materialize(
        [asset1],
        partition_key="2022-01-01",
        resources=resource_defs,
        run_config={"ops": {"my_schema__" + asset1_path: {"config": {"value": "3"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == ["2", "2", "2", "3", "3", "3"]


@asset(
    partitions_def=DailyPartitionsDefinition(start_date="2022-01-01"),
    key_prefix=["my_schema"],
    metadata={"partition_expr": "time"},
)
def load_partitioned(daily_partitioned: pl.DataFrame) -> pl.DataFrame:
    return daily_partitioned


def test_load_partitioned_asset(io_manager):
    resource_defs = {"io_manager": io_manager}

    res = materialize(
        [daily_partitioned, load_partitioned],
        partition_key="2022-01-01",
        resources=resource_defs,
        run_config={"ops": {"my_schema__daily_partitioned": {"config": {"value": "1"}}}},
    )

    assert res.success
    table = res.asset_value(["my_schema", "load_partitioned"])
    assert table.shape[0] == 3

    res = materialize(
        [daily_partitioned, load_partitioned],
        partition_key="2022-01-02",
        resources=resource_defs,
        run_config={"ops": {"my_schema__daily_partitioned": {"config": {"value": "2"}}}},
    )

    assert res.success
    table = res.asset_value(["my_schema", "load_partitioned"])
    assert table.shape[0] == 3


@asset(
    partitions_def=StaticPartitionsDefinition(["red", "yellow", "blue"]),
    key_prefix=["my_schema"],
    metadata={"partition_expr": "color"},
    config_schema={"value": str},
)
def static_partitioned(context) -> pl.DataFrame:
    partition = context.partition_key
    value = context.op_execution_context.op_config["value"]

    return pl.DataFrame(
        {
            "color": [partition, partition, partition],
            "a": [value, value, value],
            "b": [4, 5, 6],
        },
    )


def test_static_partitioned_asset(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    materialize(
        [static_partitioned],
        partition_key="red",
        resources=resource_defs,
        run_config={"ops": {"my_schema__static_partitioned": {"config": {"value": "1"}}}},
    )

    dt = DeltaTable(os.path.join(tmp_path, "my_schema/static_partitioned"))
    out_df = dt.to_pyarrow_table()
    assert out_df["a"].to_pylist() == ["1", "1", "1"]

    materialize(
        [static_partitioned],
        partition_key="blue",
        resources=resource_defs,
        run_config={"ops": {"my_schema__static_partitioned": {"config": {"value": "2"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == ["1", "1", "1", "2", "2", "2"]

    materialize(
        [static_partitioned],
        partition_key="red",
        resources=resource_defs,
        run_config={"ops": {"my_schema__static_partitioned": {"config": {"value": "3"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == ["2", "2", "2", "3", "3", "3"]


@asset(
    partitions_def=MultiPartitionsDefinition(
        {
            "time": DailyPartitionsDefinition(start_date="2022-01-01"),
            "color": StaticPartitionsDefinition(["red", "yellow", "blue"]),
        },
    ),
    key_prefix=["my_schema"],
    metadata={"partition_expr": {"time": "time", "color": "color"}},
    config_schema={"value": str},
)
def multi_partitioned(context) -> pl.DataFrame:
    partition = context.partition_key.keys_by_dimension
    time_partition = datetime.strptime(partition["time"], DELTA_DATE_FORMAT).date()
    value = context.op_execution_context.op_config["value"]
    return pl.DataFrame(
        {
            "color": [partition["color"], partition["color"], partition["color"]],
            "time": [time_partition, time_partition, time_partition],
            "a": [value, value, value],
        },
    )


def test_multi_partitioned_asset(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    materialize(
        [multi_partitioned],
        partition_key=MultiPartitionKey({"time": "2022-01-01", "color": "red"}),
        resources=resource_defs,
        run_config={"ops": {"my_schema__multi_partitioned": {"config": {"value": "1"}}}},
    )

    dt = DeltaTable(os.path.join(tmp_path, "my_schema/multi_partitioned"))
    out_df = dt.to_pyarrow_table()
    assert out_df["a"].to_pylist() == ["1", "1", "1"]

    materialize(
        [multi_partitioned],
        partition_key=MultiPartitionKey({"time": "2022-01-01", "color": "blue"}),
        resources=resource_defs,
        run_config={"ops": {"my_schema__multi_partitioned": {"config": {"value": "2"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == ["1", "1", "1", "2", "2", "2"]

    materialize(
        [multi_partitioned],
        partition_key=MultiPartitionKey({"time": "2022-01-02", "color": "red"}),
        resources=resource_defs,
        run_config={"ops": {"my_schema__multi_partitioned": {"config": {"value": "3"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == [
        "1",
        "1",
        "1",
        "2",
        "2",
        "2",
        "3",
        "3",
        "3",
    ]

    materialize(
        [multi_partitioned],
        partition_key=MultiPartitionKey({"time": "2022-01-01", "color": "red"}),
        resources=resource_defs,
        run_config={"ops": {"my_schema__multi_partitioned": {"config": {"value": "4"}}}},
    )

    dt.update_incremental()
    out_df = dt.to_pyarrow_table()
    assert sorted(out_df["a"].to_pylist()) == [
        "2",
        "2",
        "2",
        "3",
        "3",
        "3",
        "4",
        "4",
        "4",
    ]


dynamic_fruits = DynamicPartitionsDefinition(name="dynamic_fruits")


@asset(
    partitions_def=dynamic_fruits,
    key_prefix=["my_schema"],
    metadata={"partition_expr": "fruit"},
    config_schema={"value": str},
)
def dynamic_partitioned(context: AssetExecutionContext) -> pl.DataFrame:
    partition = context.partition_key
    value = context.op_execution_context.op_config["value"]
    return pl.DataFrame(
        {
            "fruit": [partition, partition, partition],
            "a": [value, value, value],
        },
    )


def test_dynamic_partition(tmp_path, io_manager):
    with instance_for_test() as instance:
        resource_defs = {"io_manager": io_manager}

        instance.add_dynamic_partitions(dynamic_fruits.name, ["apple"])  # type: ignore

        materialize(
            [dynamic_partitioned],
            partition_key="apple",
            resources=resource_defs,
            instance=instance,
            run_config={"ops": {"my_schema__dynamic_partitioned": {"config": {"value": "1"}}}},
        )

        dt = DeltaTable(os.path.join(tmp_path, "my_schema/dynamic_partitioned"))
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == ["1", "1", "1"]

        instance.add_dynamic_partitions(dynamic_fruits.name, ["orange"])  # type: ignore

        materialize(
            [dynamic_partitioned],
            partition_key="orange",
            resources=resource_defs,
            instance=instance,
            run_config={"ops": {"my_schema__dynamic_partitioned": {"config": {"value": "2"}}}},
        )

        dt.update_incremental()
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == ["1", "1", "1", "2", "2", "2"]

        materialize(
            [dynamic_partitioned],
            partition_key="apple",
            resources=resource_defs,
            instance=instance,
            run_config={"ops": {"my_schema__dynamic_partitioned": {"config": {"value": "3"}}}},
        )

        dt.update_incremental()
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == ["2", "2", "2", "3", "3", "3"]



from enum import Enum
from typing import Literal, Optional

from dagster import Config


def _to_str_dict(dictionary: dict) -> dict[str, str]:
    """Filters dict of None values and casts other values to str."""
    return {key: str(value) for key, value in dictionary.items() if value is not None}


class LocalConfig(Config):
    """Storage configuration for local object store."""

    provider: Literal["local"] = "local"

    def str_dict(self) -> dict[str, str]:
        """Storage options as str dict."""
        return _to_str_dict(self.model_dump())


class AzureConfig(Config):
    """Storage configuration for Microsoft Azure Blob or ADLS Gen 2 object store."""

    provider: Literal["azure"] = "azure"

    account_name: Optional[str] = None
    """Storage account name"""

    client_id: Optional[str] = None
    """Client ID for ID / secret based authentication."""

    client_secret: Optional[str] = None
    """Client secret for ID / secret based authentication."""

    tenant_id: Optional[str] = None
    """Tenant ID for ID / secret based authentication."""

    federated_token_file: Optional[str] = None
    """File containing federated credential token"""

    account_key: Optional[str] = None
    """Storage account master key"""

    sas_key: Optional[str] = None
    """Shared access signature"""

    token: Optional[str] = None
    """Hard-coded bearer token"""

    use_azure_cli: Optional[bool] = None
    """Use azure cli for acquiring access token"""

    use_fabric_endpoint: Optional[bool] = None
    """Use object store with url scheme account.dfs.fabric.microsoft.com"""

    msi_resource_id: Optional[str] = None
    """Msi resource id for use with managed identity authentication."""

    msi_endpoint: Optional[str] = None
    """Endpoint to request a imds managed identity token."""

    container_name: Optional[str] = None
    """Storage container name"""

    def str_dict(self) -> dict[str, str]:
        """Storage options as str dict."""
        return _to_str_dict(self.model_dump())


class S3Config(Config):
    """Storage configuration for Amazon Web Services (AWS) S3 object store."""

    provider: Literal["s3"] = "s3"

    access_key_id: Optional[str] = None
    """AWS access key ID"""

    secret_access_key: Optional[str] = None
    """AWS access key secret"""

    region: Optional[str] = None
    """AWS region"""

    bucket: Optional[str] = None
    """Storage bucket name"""

    endpoint: Optional[str] = None
    """Sets custom endpoint for communicating with S3."""

    token: Optional[str] = None
    """Token to use for requests (passed to underlying provider)"""

    imdsv1_fallback: bool = False
    """Allow fall back to ImdsV1"""

    virtual_hosted_style_request: Optional[str] = None
    """Bucket is hosted under virtual-hosted-style URL"""

    unsigned_payload: Optional[bool] = None
    """Avoid computing payload checksum when calculating signature."""

    checksum: Optional[str] = None
    """Set the checksum algorithm for this client."""

    metadata_endpoint: Optional[str] = None
    """Instance metadata endpoint URL for fetching credentials"""

    container_credentials_relative_uri: Optional[str] = None
    """Set the container credentials relative URI

    https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-iam-roles.html
    """

    def str_dict(self) -> dict[str, str]:
        """Storage options as str dict."""
        return _to_str_dict(self.model_dump())


class GcsConfig(Config):
    """Storage configuration for Google Cloud Storage object store."""

    provider: Literal["gcs"] = "gcs"

    service_account: Optional[str] = None
    """Path to the service account file"""

    service_account_key: Optional[str] = None
    """The serialized service account key."""

    bucket: Optional[str] = None
    """Bucket name"""

    application_credentials: Optional[str] = None
    """Application credentials path"""

    def str_dict(self) -> dict[str, str]:
        """Storage options as str dict."""
        return _to_str_dict(self.model_dump())


class BackoffConfig(Config):
    """Configuration for exponential back off https://docs.rs/object_store/latest/object_store/struct.BackoffConfig.html"""

    init_backoff: Optional[str] = None
    """The initial backoff duration"""

    max_backoff: Optional[str] = None
    """The maximum backoff duration"""

    base: Optional[float] = None
    """The multiplier to use for the next backoff duration"""


class ClientConfig(Config):
    """Configuration for http client interacting with storage APIs."""

    allow_http: Optional[bool] = None
    """Allow non-TLS, i.e. non-HTTPS connections"""

    allow_invalid_certificates: Optional[bool] = None
    """Skip certificate validation on https connections.

    ## Warning

    You should think very carefully before using this method.
    If invalid certificates are trusted, any certificate for any site will be trusted for use.
    This includes expired certificates. This introduces significant vulnerabilities,
    and should only be used as a last resort or for testing
    """

    connect_timeout: Optional[str] = None
    """Timeout for only the connect phase of a Client"""

    default_content_type: Optional[str] = None
    """default CONTENT_TYPE for uploads"""

    http1_only: Optional[bool] = None
    """Only use http1 connections"""

    http2_keep_alive_interval: Optional[int] = None
    """Interval for HTTP2 Ping frames should be sent to keep a connection alive."""

    http2_keep_alive_timeout: Optional[int] = None
    """Timeout for receiving an acknowledgement of the keep-alive ping."""

    http2_keep_alive_while_idle: Optional[int] = None
    """Enable HTTP2 keep alive pings for idle connections"""

    http2_only: Optional[bool] = None
    """Only use http2 connections"""

    pool_idle_timeout: Optional[str] = None
    """The pool max idle timeout

    This is the length of time an idle connection will be kept alive
    """

    pool_max_idle_per_host: Optional[int] = None
    """maximum number of idle connections per host"""

    proxy_url: Optional[str] = None
    """HTTP proxy to use for requests"""

    timeout: Optional[str] = None
    """Request timeout, e.g. 10s, 60s

    The timeout is applied from when the request starts connecting until the response body has finished
    """

    user_agent: Optional[str] = None
    """User-Agent header to be used by this client"""

    OBJECT_STORE_CONCURRENCY_LIMIT: Optional[str] = None
    """The number of concurrent connections the underlying object store can create"""

    MOUNT_ALLOW_UNSAFE_RENAME: Optional[str] = None
    """If set it will allow unsafe renames on mounted storage"""

    max_retries: Optional[int] = None
    """The maximum number of times to retry a request. Set to 0 to disable retries"""

    retry_timeout: Optional[str] = None
    """The maximum duration of time from the initial request after which no further retries will be attempted. e.g. 10s, 60s"""

    backoff_config: Optional[BackoffConfig] = None
    """Configuration for exponential back off """

    def str_dict(self) -> dict[str, str]:
        """Storage options as str dict."""
        model_dump = self.model_dump()
        str_dict: dict[str, str] = {}
        for key, value in model_dump.items():
            if value is not None:
                if isinstance(value, BackoffConfig):
                    ## delta-rs uses custom config keys for the BackOffConfig
                    ## https://delta-io.github.io/delta-rs/integrations/object-storage/special_configuration/
                    if value.base is not None:
                        str_dict["backoff_config.base"] = str(value.base)
                    if value.max_backoff is not None:
                        str_dict["backoff_config.max_backoff"] = str(value.max_backoff)
                    if value.init_backoff is not None:
                        str_dict["backoff_config.init_backoff"] = str(value.init_backoff)
                else:
                    str_dict[key] = str(value)
        return str_dict


class MergeType(str, Enum):
    """Enum of the possible IO Manager merge types
    - "deduplicate_insert"  <- Deduplicates on write
    - "update_only"  <- updates only the matches records
    - "upsert"  <- updates existing matches and inserts non matched records
    - "replace_and_delete_unmatched" <- updates existing matches and deletes unmatched
    - "custom" <- requires MergeOperationsConfig to be provided
    """

    deduplicate_insert = "deduplicate_insert"  # Deduplicates on write
    update_only = "update_only"  # updates only the records
    upsert = "upsert"  # updates and inserts
    replace_delete_unmatched = "replace_and_delete_unmatched"
    custom = "custom"


class OperationConfig(Config):
    """Basic operation config"""

    predicate: Optional[str] = None


class OperationAllConfig(Config):
    """Configuration for `_all` operations"""

    predicate: Optional[str] = None
    except_cols: Optional[list[str]] = None


class OperationWithUpdatesConfig(Config):
    """Configuration for operations that allow specific column updates"""

    updates: dict[str, str]
    predicate: Optional[str] = None


class WhenNotMatchedInsert(OperationWithUpdatesConfig):
    """When not matched statement"""

    pass


class WhenNotMatchedInsertAll(OperationAllConfig):
    """When not matched insert all statement"""

    pass


class WhenMatchedUpdate(OperationWithUpdatesConfig):
    """When matched update statement"""

    pass


class WhenMatchedUpdateAll(OperationAllConfig):
    """When matched update all statement"""

    pass


class WhenMatchedDelete(OperationConfig):
    """When matched delete statement"""

    pass


class WhenNotMatchedBySourceDelete(OperationConfig):
    """When not matched by source delete statement"""

    pass


class WhenNotMatchedBySourceUpdate(OperationWithUpdatesConfig):
    """When not matched by source update statement"""

    pass


class MergeOperationsConfig(Config):
    """Configuration for each merge operation. Only used with merge_type 'custom'.

    If you have multiple when statements of a single operation, they are evaluated in the order as provided in the list.
    """

    when_not_matched_insert: Optional[list[WhenNotMatchedInsert]] = None
    when_not_matched_insert_all: Optional[list[WhenNotMatchedInsertAll]] = None
    when_matched_update: Optional[list[WhenMatchedUpdate]] = None
    when_matched_update_all: Optional[list[OperationAllConfig]] = None
    when_matched_delete: Optional[list[WhenMatchedDelete]] = None
    when_not_matched_by_source_delete: Optional[list[WhenNotMatchedBySourceDelete]] = None
    when_not_matched_by_source_update: Optional[list[WhenNotMatchedBySourceUpdate]] = None


class MergeConfig(Config):
    """Configuration for the MERGE operation."""

    merge_type: MergeType
    """The type of MERGE to execute."""

    predicate: Optional[str] = None
    """SQL like predicate on how to merge, passed into DeltaTable.merge()

    This can also be set on the asset definition metadata using the `merge_predicate` key"""

    source_alias: Optional[str] = None
    """Alias for the source table"""

    target_alias: Optional[str] = None
    """Alias for the target table"""

    error_on_type_mismatch: bool = True
    """specify if merge will return error if data types are mismatching"""

    merge_operations_config: Optional[MergeOperationsConfig] = None
    """Full configuration of each merge operation, only use with merge_type='custom'"""



import os

import pyarrow as pa
from dagster import (
    asset,
    materialize,
)
from deltalake import DeltaTable

from dagster_delta import DeltaLakePyarrowIOManager


@asset(
    key_prefix=["my_schema"],
    metadata={"root_name": "custom_asset"},
)
def asset_1() -> pa.Table:
    return pa.Table.from_pydict(
        {
            "value": [1],
            "b": [1],
        },
    )


def test_asset_with_root_name(
    tmp_path,
    io_manager: DeltaLakePyarrowIOManager,
):
    resource_defs = {"io_manager": io_manager}

    res = materialize([asset_1], resources=resource_defs)
    assert res.success

    data = res.asset_value(asset_1.key)

    dt = DeltaTable(os.path.join(str(tmp_path), "my_schema", "custom_asset"))

    assert data == dt.to_pyarrow_table()



import sys
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, TypedDict, Union, cast

from dagster import InputContext, OutputContext
from dagster._config.pythonic_config import ConfigurableIOManagerFactory

try:
    from dagster._core.definitions.partitions.utils import TimeWindow
except ModuleNotFoundError:
    from dagster._core.definitions.time_window_partitions import TimeWindow

from dagster._core.storage.db_io_manager import (
    DbClient,
    DbTypeHandler,
    TablePartitionDimension,
    TableSlice,
)
from pydantic import Field

from dagster_delta._db_io_manager import CustomDbIOManager

if sys.version_info >= (3, 11):
    from typing import NotRequired
else:
    from typing_extensions import NotRequired

from dagster_delta.config import (
    AzureConfig,
    ClientConfig,
    GcsConfig,
    LocalConfig,
    MergeConfig,
    S3Config,
)

DELTA_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
DELTA_DATE_FORMAT = "%Y-%m-%d"


@dataclass(frozen=True)
class TableConnection:  # noqa: D101
    table_uri: str
    storage_options: dict[str, str]
    table_config: Optional[dict[str, str]]


class _StorageOptionsConfig(TypedDict, total=False):
    local: dict[str, str]
    s3: dict[str, str]
    azure: dict[str, str]
    gcs: dict[str, str]


class WriteMode(str, Enum):
    """Deltalake Write mode"""

    error = "error"
    append = "append"
    overwrite = "overwrite"
    ignore = "ignore"
    merge = "merge"
    create_or_replace = "create_or_replace"


class SchemaMode(str, Enum):
    """Deltalake schema mode"""

    overwrite = "overwrite"
    merge = "merge"


class _DeltaTableIOManagerResourceConfig(TypedDict):
    root_uri: str
    mode: WriteMode
    schema_mode: NotRequired[SchemaMode]
    merge_config: NotRequired[dict[str, Any]]
    storage_options: _StorageOptionsConfig
    client_options: NotRequired[dict[str, str]]
    table_config: NotRequired[dict[str, str]]
    writer_properties: NotRequired[dict[str, str]]
    commit_properties: NotRequired[dict[str, str]]
    parquet_read_options: NotRequired[dict[str, Any]]


class BaseDeltaLakeIOManager(ConfigurableIOManagerFactory):
    """Base class for an IO manager definition that reads inputs from and writes outputs to Delta Lake.

    Examples:
    ```python
    from dagster_delta import BaseDeltaLakeIOManager
    from dagster_delta.io_manager.arrow import _DeltaLakePyArrowTypeHandler

    class MyDeltaLakeIOManager(BaseDeltaLakeIOManager):
        @staticmethod
        def type_handlers() -> Sequence[DbTypeHandler]:
            return [_DeltaLakePyArrowTypeHandler()]

    @asset(
        key_prefix=["my_schema"]  # will be used as the schema (parent folder) in Delta Lake
    )
    def my_table() -> pa.Table:  # the name of the asset will be the table name
        ...

    defs = Definitions(
        assets=[my_table],
        resources={"io_manager": MyDeltaLakeIOManager()}
    )
    ```

    If you do not provide a schema, Dagster will determine a schema based on the assets and ops using
    the I/O Manager. For assets, the schema will be determined from the asset key, as in the above example.
    For ops, the schema can be specified by including a "schema" entry in output metadata. If none
    of these is provided, the schema will default to "public".

    ```python

    @op(
        out={"my_table": Out(metadata={"schema": "my_schema"})}
    )
    def make_my_table() -> pa.Table:
        ...
    ```

    To only use specific columns of a table as input to a downstream op or asset, add the metadata "columns" to the
    In or AssetIn.

    ```python
        @asset(
            ins={"my_table": AssetIn("my_table", metadata={"columns": ["a"]})}
        )
        def my_table_a(my_table: pa.Table:
            # my_table will just contain the data from column "a"
            ...
    ```
    """

    root_uri: str = Field(description="Storage location where Delta tables are stored.")
    mode: WriteMode = Field(  # type: ignore
        default=WriteMode.overwrite.value,
        description="The write mode passed to save the output.",
    )
    schema_mode: Optional[SchemaMode] = Field(
        default=None,
        description="If set to 'overwrite', allows replacing the schema of the table. Set to 'merge' to merge with existing schema.",
    )

    merge_config: Optional[MergeConfig] = Field(
        default=None,
        description="Additional configuration for the MERGE operation.",
    )

    storage_options: Union[AzureConfig, S3Config, LocalConfig, GcsConfig] = Field(  # noqa
        discriminator="provider",
        description="Configuration for accessing storage location.",
    )

    client_options: Optional[ClientConfig] = Field(
        default=None,
        description="Additional configuration passed to http client.",
    )

    table_config: Optional[dict[str, str]] = Field(
        default=None,
        description="Additional config and metadata added to table on creation.",
    )

    schema_: Optional[str] = Field(
        default=None,
        alias="schema",
        description="Name of the schema to use.",
    )  # schema is a reserved word for pydantic
    writer_properties: Optional[dict[str, str]] = Field(
        default=None,
        description="Writer properties passed to the rust engine writer.",
    )
    commit_properties: Optional[dict[str, Any]] = Field(
        default=None,
        description="Commit properties, Controls the behaviour of the commit.",
    )
    parquet_read_options: Optional[dict[str, Any]] = Field(
        default=None,
        description="Parquet read options, to be passed to pyarrow.",
    )

    @staticmethod
    @abstractmethod
    def type_handlers() -> Sequence[DbTypeHandler]:  # noqa: D102
        ...

    @staticmethod
    def default_load_type() -> Optional[type]:  # noqa: D102
        return None

    def create_io_manager(self, context) -> CustomDbIOManager:  # noqa: D102, ANN001, ARG002
        self.storage_options.model_dump()
        return CustomDbIOManager(
            db_client=DeltaLakeDbClient(),
            database="deltalake",
            schema=self.schema_,
            type_handlers=self.type_handlers(),
            default_load_type=self.default_load_type(),
            io_manager_name="DeltaLakeIOManager",
        )


class DeltaLakeDbClient(DbClient):  # noqa: D101
    @staticmethod
    def delete_table_slice(  # noqa: D102
        context: OutputContext,
        table_slice: TableSlice,
        connection: TableConnection,
    ) -> None:
        # deleting the table slice here is a no-op, since we use deltalake's internal mechanism
        # to overwrite table partitions.
        pass

    @staticmethod
    def ensure_schema_exists(  # noqa: D102
        context: OutputContext,
        table_slice: TableSlice,
        connection: TableConnection,
    ) -> None:
        # schemas are just folders and automatically created on write.
        pass

    @staticmethod
    def get_select_statement(table_slice: TableSlice) -> str:  # noqa: D102
        # The select statement here is just for illustrative purposes,
        # and is never actually executed. It does however logically correspond
        # the the operation being executed.
        col_str = ", ".join(table_slice.columns) if table_slice.columns else "*"

        if table_slice.partition_dimensions and len(table_slice.partition_dimensions) > 0:
            query = f"SELECT {col_str} FROM {table_slice.schema}.{table_slice.table} WHERE\n"
            return query + _partition_where_clause(table_slice.partition_dimensions)
        else:
            return f"""SELECT {col_str} FROM {table_slice.schema}.{table_slice.table}"""

    @staticmethod
    @contextmanager
    def connect(  # noqa: D102
        context: Union[OutputContext, InputContext],
        table_slice: TableSlice,
    ) -> Iterator[TableConnection]:
        resource_config = cast(_DeltaTableIOManagerResourceConfig, context.resource_config)
        root_uri = resource_config["root_uri"].rstrip("/")
        storage_options = resource_config["storage_options"]

        # Values of a config are unpacked into a dict[str,Any], we convert it back to the Config
        # so that we can do str_dict()
        if "local" in storage_options:
            storage_options = LocalConfig(**storage_options["local"])  # type: ignore
        elif "s3" in storage_options:
            storage_options = S3Config(**storage_options["s3"])  # type: ignore
        elif "azure" in storage_options:
            storage_options = AzureConfig(**storage_options["azure"])  # type: ignore
        elif "gcs" in storage_options:
            storage_options = GcsConfig(**storage_options["gcs"])  # type: ignore
        else:
            raise NotImplementedError("No valid storage_options config found.")

        client_options = resource_config.get("client_options")
        client_options = (
            ClientConfig(**client_options).str_dict() if client_options is not None else {}  # type: ignore
        )

        options = {**storage_options.str_dict(), **client_options}
        table_config = resource_config.get("table_config")

        # Ignore schema if None or empty string, useful to set schema = "" which overrides the assetkey
        if table_slice.schema:
            table_uri = f"{root_uri}/{table_slice.schema}/{table_slice.table}"
        else:
            table_uri = f"{root_uri}/{table_slice.table}"
        conn = TableConnection(
            table_uri=table_uri,
            storage_options=options or {},
            table_config=table_config,
        )

        yield conn


def _partition_where_clause(
    partition_dimensions: Sequence[TablePartitionDimension],
) -> str:
    return " AND\n".join(
        (
            _time_window_where_clause(partition_dimension)
            if isinstance(partition_dimension.partitions, TimeWindow)
            else _static_where_clause(partition_dimension)
        )
        for partition_dimension in partition_dimensions
    )


def _time_window_where_clause(table_partition: TablePartitionDimension) -> str:
    partition = cast(TimeWindow, table_partition.partitions)
    start_dt, end_dt = partition
    start_dt_str = start_dt.strftime(DELTA_DATETIME_FORMAT)
    end_dt_str = end_dt.strftime(DELTA_DATETIME_FORMAT)
    return f"""{table_partition.partition_expr} >= '{start_dt_str}' AND {table_partition.partition_expr} < '{end_dt_str}'"""


def _static_where_clause(table_partition: TablePartitionDimension) -> str:
    partitions = ", ".join(f"'{partition}'" for partition in table_partition.partitions)
    return f"""{table_partition.partition_expr} in ({partitions})"""



from collections.abc import Mapping
from typing import (
    TypeVar,
    Union,
    cast,
)

try:
    from dagster._core.definitions.partitions.definition.multi import (
        MultiPartitionsDefinition,
    )
    from dagster._core.definitions.partitions.definition.time_window import (
        TimeWindowPartitionsDefinition,
    )
except ModuleNotFoundError:
    from dagster._core.definitions.multi_dimensional_partitions import (
        MultiPartitionsDefinition,
    )
    from dagster._core.definitions.time_window_partitions import (
        TimeWindowPartitionsDefinition,
    )

from dagster._core.execution.context.input import InputContext
from dagster._core.execution.context.output import OutputContext
from dagster._core.storage.db_io_manager import DbIOManager, TablePartitionDimension, TableSlice

from dagster_delta._db_io_manager.utils import (
    generate_multi_partitions_dimension,
    generate_single_partition_dimension,
)

T = TypeVar("T")


class CustomDbIOManager(DbIOManager):
    """Works exactly like the DbIOManager, but overrides the _get_table_slice method
    to provide support for partition mapping. e.g. a mapping from partition A to partition B,
    where A is partitioned on two dimensions and B is partitioned on only one dimension.

    Additionally, gives ability to override
    the table name using `root_name` in the metadata.

    Example:
    ```
    @dg.asset(
    partitions_def=dg.StaticPartitionsDefinition(["a", "b"]),
    metadata={
        "partition_expr": "foo",
        "root_name": "asset_partitioned",
        },
    )
    def asset_partitioned_1(upstream_1, upstream_2):
    ```
    """

    def _get_table_slice(
        self,
        context: Union[OutputContext, InputContext],
        output_context: OutputContext,
    ) -> TableSlice:
        output_context_definition_metadata = output_context.definition_metadata or {}

        schema: str
        table: str
        partition_dimensions: list[TablePartitionDimension] = []
        if context.has_asset_key:
            asset_key_path = context.asset_key.path

            ## Override the
            if output_context_definition_metadata.get("root_name"):
                table = output_context_definition_metadata["root_name"]
            else:
                table = asset_key_path[-1]
            # schema order of precedence: metadata, I/O manager 'schema' config, key_prefix
            if output_context_definition_metadata.get("schema"):
                schema = cast(str, output_context_definition_metadata["schema"])
            elif self._schema:
                schema = self._schema
            elif len(asset_key_path) > 1:
                schema = asset_key_path[-2]
            else:
                schema = "public"

            if context.has_asset_partitions:
                partition_expr = output_context_definition_metadata.get("partition_expr")
                if partition_expr is None:
                    raise ValueError(
                        f"Asset '{context.asset_key}' has partitions, but no 'partition_expr'"
                        " metadata value, so we don't know what column it's partitioned on. To"
                        " specify a column, set this metadata value. E.g."
                        ' @asset(metadata={"partition_expr": "your_partition_column"}).',
                    )

                if isinstance(context.asset_partitions_def, MultiPartitionsDefinition):
                    partition_dimensions.extend(
                        generate_multi_partitions_dimension(
                            asset_partition_keys=context.asset_partition_keys,
                            asset_partitions_def=context.asset_partitions_def,
                            partition_expr=cast(Mapping[str, str], partition_expr),
                            asset_key=context.asset_key,
                        ),
                    )
                else:
                    partition_dimensions.append(
                        generate_single_partition_dimension(
                            partition_expr=cast(str, partition_expr),
                            asset_partition_keys=context.asset_partition_keys,
                            asset_partitions_time_window=(
                                context.asset_partitions_time_window
                                if isinstance(
                                    context.asset_partitions_def,
                                    TimeWindowPartitionsDefinition,
                                )
                                else None
                            ),
                        ),
                    )
        else:
            table = output_context.name
            if output_context_definition_metadata.get("schema"):
                schema = cast(str, output_context_definition_metadata["schema"])
            elif self._schema:
                schema = self._schema
            else:
                schema = "public"

        return TableSlice(
            table=table,
            schema=schema,
            database=self._database,
            partition_dimensions=partition_dimensions,
            columns=(context.definition_metadata or {}).get("columns"),
        )



import pytest
from dagster import (
    DailyPartitionsDefinition,
    MultiPartitionsDefinition,
    StaticPartitionsDefinition,
)

from dagster_delta import DeltaLakePyarrowIOManager
from dagster_delta.config import LocalConfig


@pytest.fixture
def io_manager(
    tmp_path,
) -> DeltaLakePyarrowIOManager:
    return DeltaLakePyarrowIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
    )


@pytest.fixture
def daily_partitions_definition() -> DailyPartitionsDefinition:
    return DailyPartitionsDefinition(start_date="2022-01-01", end_date="2022-01-10")


@pytest.fixture
def letter_partitions_definition() -> StaticPartitionsDefinition:
    return StaticPartitionsDefinition(["a", "b", "c"])


@pytest.fixture
def color_partitions_definition() -> StaticPartitionsDefinition:
    return StaticPartitionsDefinition(["red", "blue", "yellow"])


@pytest.fixture
def multi_partition_with_letter(
    daily_partitions_definition: DailyPartitionsDefinition,
    letter_partitions_definition: StaticPartitionsDefinition,
) -> MultiPartitionsDefinition:
    return MultiPartitionsDefinition(
        partitions_defs={
            "date": daily_partitions_definition,
            "letter": letter_partitions_definition,
        },
    )


@pytest.fixture
def multi_partition_with_color(
    daily_partitions_definition: DailyPartitionsDefinition,
    color_partitions_definition: StaticPartitionsDefinition,
) -> MultiPartitionsDefinition:
    return MultiPartitionsDefinition(
        partitions_defs={
            "date": daily_partitions_definition,
            "color": color_partitions_definition,
        },
    )



from dagster_delta.config import (
    AzureConfig,
    BackoffConfig,
    ClientConfig,
    GcsConfig,
    LocalConfig,
    MergeConfig,
    MergeOperationsConfig,
    MergeType,
    S3Config,
    WhenMatchedDelete,
    WhenMatchedUpdate,
    WhenMatchedUpdateAll,
    WhenNotMatchedBySourceDelete,
    WhenNotMatchedBySourceUpdate,
    WhenNotMatchedInsert,
    WhenNotMatchedInsertAll,
)
from dagster_delta.io_manager.arrow import DeltaLakePyarrowIOManager
from dagster_delta.io_manager.base import (
    BaseDeltaLakeIOManager,
    SchemaMode,
    WriteMode,
)
from dagster_delta.resources import DeltaTableResource

__all__ = [
    "AzureConfig",
    "ClientConfig",
    "GcsConfig",
    "S3Config",
    "LocalConfig",
    "BackoffConfig",
    "MergeConfig",
    "MergeType",
    "WriteMode",
    "SchemaMode",
    "DeltaTableResource",
    "BaseDeltaLakeIOManager",
    "DeltaLakePyarrowIOManager",
    "WhenMatchedDelete",
    "WhenMatchedUpdate",
    "WhenMatchedUpdateAll",
    "WhenNotMatchedBySourceDelete",
    "WhenNotMatchedBySourceUpdate",
    "WhenNotMatchedInsert",
    "WhenNotMatchedInsertAll",
    "MergeOperationsConfig",
]


try:
    from dagster_delta.io_manager.polars import DeltaLakePolarsIOManager  # noqa

    __all__.extend(["DeltaLakePolarsIOManager"])

except ImportError as e:
    if "polars" in str(e):
        pass
    else:
        raise e



from datetime import date, datetime
from typing import Optional

from deltalake.table import FilterLiteralType

from dagster_delta.io_manager.base import (
    DELTA_DATE_FORMAT,
    DELTA_DATETIME_FORMAT,
)


def create_predicate(
    partition_filters: list[FilterLiteralType],
    target_alias: Optional[str] = None,
) -> str:
    partition_predicates = []
    for part_filter in partition_filters:
        column = f"{target_alias}.{part_filter[0]}" if target_alias is not None else part_filter[0]
        value = part_filter[2]
        if isinstance(value, (int, float, bool)):
            value = str(value)
        elif isinstance(value, str):
            value = f"'{value}'"
        elif isinstance(value, list):
            value = str(tuple(v for v in value))
        elif isinstance(value, date):
            value = f"'{value.strftime(DELTA_DATE_FORMAT)}'"
        elif isinstance(value, datetime):
            value = f"'{value.strftime(DELTA_DATETIME_FORMAT)}'"
        partition_predicates.append(f"{column} {part_filter[1]} {value}")

    return " AND ".join(partition_predicates)



import os

import pyarrow as pa
import pytest
from dagster import (
    OpExecutionContext,
    Out,
    graph,
    op,
)
from deltalake import DeltaTable

from dagster_delta import (
    DeltaLakePyarrowIOManager,
    LocalConfig,
    MergeConfig,
    MergeOperationsConfig,
    MergeType,
    WriteMode,
)
from dagster_delta.config import WhenMatchedUpdateAll, WhenNotMatchedInsertAll


@op(out=Out(metadata={"schema": "a_df2"}))
def a_df2(context: OpExecutionContext) -> pa.Table:
    context.add_output_metadata(
        {
            "merge_predicate": "s.a = t.a",
            "merge_operations_config": MergeOperationsConfig(
                when_not_matched_insert_all=[WhenNotMatchedInsertAll()],
                when_matched_update_all=[WhenMatchedUpdateAll()],
            ).model_dump(),
        },
    )

    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one2"}))
def add_one2(context: OpExecutionContext, df: pa.RecordBatchReader):  # noqa: ANN201
    context.add_output_metadata(
        {
            "merge_predicate": "s.a = t.a",
            "merge_operations_config": MergeOperationsConfig(
                when_not_matched_insert_all=[WhenNotMatchedInsertAll()],
                when_matched_update_all=[WhenMatchedUpdateAll()],
            ).model_dump(),
        },
    )

    return df.read_all().set_column(0, "a", pa.array([2, 3, 4]))


@graph
def add_one_to_dataframe_2():
    add_one2(a_df2())


@op(out=Out(metadata={"schema": "a_df"}))
def a_df() -> pa.Table:
    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@op(out=Out(metadata={"schema": "add_one"}))
def add_one(df: pa.Table):  # noqa: ANN201
    return df.set_column(0, "a", pa.array([2, 3, 4]))


@graph
def add_one_to_dataframe():
    add_one(a_df())


@pytest.mark.parametrize("merge_type", [MergeType.deduplicate_insert, MergeType.upsert])
def test_deltalake_io_manager_with_ops_rust_writer(tmp_path, merge_type: MergeType):
    resource_defs = {
        "io_manager": DeltaLakePyarrowIOManager(
            root_uri=str(tmp_path),
            storage_options=LocalConfig(),
            mode=WriteMode.merge,
            merge_config=MergeConfig(
                merge_type=merge_type,
                predicate="s.a = t.a",
                source_alias="s",
                target_alias="t",
            ),
        ),
    }

    job = add_one_to_dataframe.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [2, 3, 4]


def test_deltalake_io_manager_custom_merge(tmp_path):
    resource_defs = {
        "io_manager": DeltaLakePyarrowIOManager(
            root_uri=str(tmp_path),
            storage_options=LocalConfig(),
            mode=WriteMode.merge,
            merge_config=MergeConfig(
                merge_type=MergeType.custom,
                predicate="s.a = t.a",
                source_alias="s",
                target_alias="t",
                merge_operations_config=MergeOperationsConfig(
                    when_not_matched_insert_all=[WhenNotMatchedInsertAll()],
                    when_matched_update_all=[WhenMatchedUpdateAll()],
                ),
            ),
        ),
    }

    job = add_one_to_dataframe.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [2, 3, 4]


def test_deltalake_io_manager_runtime_metadata_merge_configuration(tmp_path):
    """Whether runtime metadata 'merge_predicate' and 'merge_operations_config' gets picked up."""
    resource_defs = {
        "io_manager": DeltaLakePyarrowIOManager(
            root_uri=str(tmp_path),
            storage_options=LocalConfig(),
            mode=WriteMode.merge,
            merge_config=MergeConfig(
                merge_type=MergeType.custom,
                source_alias="s",
                target_alias="t",
            ),
        ),
    }

    job = add_one_to_dataframe_2.to_job(resource_defs=resource_defs)

    # run the job twice to ensure that tables get properly deleted
    for _ in range(2):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df2/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [1, 2, 3]

        dt = DeltaTable(os.path.join(tmp_path, "add_one2/result"))
        out_df = dt.to_pyarrow_table()
        assert sorted(out_df["a"].to_pylist()) == [2, 3, 4]



import logging
from collections.abc import Sequence
from typing import Optional, Union

from deltalake import DeltaTable

from dagster_delta._handler.utils import (
    create_predicate,
    extract_date_format_from_partition_definition,
    partition_dimensions_to_dnf,
)

try:
    import polars as pl
except ImportError as e:
    raise ImportError(
        "Please install dagster-delta[polars]",
    ) from e
from arro3.core import RecordBatchReader, Table
from dagster import InputContext, MetadataValue, OutputContext
from dagster._core.storage.db_io_manager import (
    DbTypeHandler,
    TableSlice,
)

from dagster_delta._handler.base import (
    DeltalakeBaseArrowTypeHandler,
)
from dagster_delta.io_manager.arrow import _DeltaLakePyArrowTypeHandler
from dagster_delta.io_manager.base import BaseDeltaLakeIOManager, TableConnection

PolarsTypes = Union[pl.DataFrame, pl.LazyFrame]


class _DeltaLakePolarsTypeHandler(DeltalakeBaseArrowTypeHandler[PolarsTypes]):  # noqa: D101
    def from_arrow(  # noqa: D102
        self,
        obj: Union[RecordBatchReader, Table],
        target_type: type[PolarsTypes],
    ) -> PolarsTypes:
        raise NotImplementedError

    def to_arrow(self, obj: PolarsTypes) -> RecordBatchReader:  # noqa: D102
        if isinstance(obj, pl.LazyFrame):
            obj = obj.collect()

        logger = logging.getLogger()
        logger.setLevel("DEBUG")
        logger.debug("shape of dataframe: %s", obj.shape)
        # TODO(ion): maybe move stats collection here

        return RecordBatchReader.from_arrow(obj)

    def load_input(
        self,
        context: InputContext,
        table_slice: TableSlice,
        connection: TableConnection,
    ) -> PolarsTypes:
        """Loads the input as a Polars DataFrame or LazyFrame."""
        definition_metadata = (
            context.definition_metadata if context.definition_metadata is not None else {}
        )
        version = definition_metadata.get("table_version")
        table = DeltaTable(
            table_uri=connection.table_uri,
            storage_options=connection.storage_options,
            version=version,
        )

        date_format = extract_date_format_from_partition_definition(context)
        logger = logging.getLogger()
        logger.setLevel("DEBUG")

        predicate = None
        if table_slice.partition_dimensions is not None:
            partition_filters = partition_dimensions_to_dnf(
                partition_dimensions=table_slice.partition_dimensions,
                table_schema=table.schema(),
                input_dnf=True,
                date_format=date_format,
            )
            if partition_filters is not None:
                ## Convert partition_filter to predicate
                predicate = create_predicate(partition_filters)

                logger.debug("Dataset input predicate %s", predicate)

        col_select = ",".join(table_slice.columns) if table_slice.columns is not None else "*"
        query = f"SELECT {col_select} FROM tbl"
        if predicate is not None:
            query = f"{query} WHERE {predicate}"

        logger.info("using query: %s", query)

        df = pl.scan_delta(table).sql(query=query, table_name="tbl")

        if context.dagster_type.typing_type == pl.LazyFrame:
            return df
        else:
            return df.collect()

    def handle_output(
        self,
        context: OutputContext,
        table_slice: TableSlice,
        obj: Union[pl.DataFrame, pl.LazyFrame],
        connection: TableConnection,
    ):
        """Writes polars frame as delta table"""
        super().handle_output(context, table_slice, obj, connection)
        metadata = {**context.consume_logged_metadata()}

        if connection.table_uri.startswith("lakefs://"):
            # We grab the lakefs endpoint from the object storage options
            for key, value in connection.storage_options.items():
                if key.lower() in ["endpoint", "aws_endpoint", "aws_endpoint_url", "endpoint_url"]:
                    metadata["lakefs_link"] = MetadataValue.url(
                        _convert_uri_to_lakefs_link(connection.table_uri, value),
                    )
                    break
        context.add_output_metadata(metadata)

    def get_output_stats(self, obj: PolarsTypes) -> dict[str, MetadataValue]:
        """Returns output stats to be attached to the the context.

        Args:
            obj (PolarsTypes): LazyFrame or DataFrame

        Returns:
            Mapping[str, MetadataValue]: metadata stats
        """
        stats = {}
        # TODO(ion): think of more meaningful stats to add from a dataframe
        if isinstance(obj, pl.DataFrame):
            stats["num_rows_in_source"] = MetadataValue.int(obj.shape[0])

        return stats

    @property
    def supported_types(self) -> Sequence[type[object]]:
        """Returns the supported dtypes for this typeHandler"""
        return [pl.DataFrame, pl.LazyFrame]


class DeltaLakePolarsIOManager(BaseDeltaLakeIOManager):
    """Base class for an IO manager definition that reads inputs from and writes outputs to Delta Lake.

    Examples:
        .. code-block:: python

            from dagster_delta import DeltaLakePolarsIOManager

            @asset(
                key_prefix=["my_schema"]  # will be used as the schema (parent folder) in Delta Lake
            )
            def my_table() -> pl.DataFrame:  # the name of the asset will be the table name
                ...

            defs = Definitions(
                assets=[my_table],
                resources={"io_manager": DeltaLakePolarsIOManager()}
            )

    If you do not provide a schema, Dagster will determine a schema based on the assets and ops using
    the I/O Manager. For assets, the schema will be determined from the asset key, as in the above example.
    For ops, the schema can be specified by including a "schema" entry in output metadata. If none
    of these is provided, the schema will default to "public".

    .. code-block:: python

        @op(
            out={"my_table": Out(metadata={"schema": "my_schema"})}
        )
        def make_my_table() -> pl.DataFrame:
            ...

    To only use specific columns of a table as input to a downstream op or asset, add the metadata "columns" to the
    In or AssetIn.

    .. code-block:: python

        @asset(
            ins={"my_table": AssetIn("my_table", metadata={"columns": ["a"]})}
        )
        def my_table_a(my_table: pl.DataFrame):
            # my_table will just contain the data from column "a"
            ...

    """

    @staticmethod
    def type_handlers() -> Sequence[DbTypeHandler]:
        """Returns all available type handlers on this IO Manager."""
        return [_DeltaLakePolarsTypeHandler(), _DeltaLakePyArrowTypeHandler()]

    @staticmethod
    def default_load_type() -> Optional[type]:
        """Grabs the default load type if no type hint is passed."""
        return pl.DataFrame


def _convert_uri_to_lakefs_link(uri: str, lakefs_base_url: str) -> str:
    """Convert an S3 uri to a link to lakefs"""
    from urllib.parse import quote

    uri = uri[len("lakefs://") :]
    parts = uri.split("/", 2)
    if len(parts) < 3:
        return "https://error-invalid-s3-uri-format"
    repository = parts[0]
    ref = parts[1]
    path = parts[2]
    encoded_path = quote(path + "/")
    https_url = f"{lakefs_base_url.rstrip('/')}/repositories/{repository}/objects?ref={ref}&path={encoded_path}"
    return https_url



import os

import pyarrow as pa
from dagster import asset, materialize
from deltalake import write_deltalake

from dagster_delta import BackoffConfig, ClientConfig, DeltaTableResource, LocalConfig


def test_resource(tmp_path):
    data = pa.table(
        {
            "a": pa.array([1, 2, 3], type=pa.int32()),
            "b": pa.array([5, 6, 7], type=pa.int32()),
        },
    )

    @asset
    def create_table(delta_table: DeltaTableResource):
        write_deltalake(
            delta_table.url,
            data,
            storage_options=delta_table.storage_options.model_dump(),
        )

    @asset
    def read_table(delta_table: DeltaTableResource):
        res = delta_table.load().to_pyarrow_table()
        assert res.equals(data)

    materialize(
        [create_table, read_table],
        resources={
            "delta_table": DeltaTableResource(
                url=os.path.join(tmp_path, "table"),
                storage_options=LocalConfig(),
                client_options=ClientConfig(
                    max_retries=10,
                    retry_timeout="10s",
                    backoff_config=BackoffConfig(init_backoff="10s", base=1.2),
                ),
            ),
        },
    )


def test_resource_versioned(tmp_path):
    data = pa.table(
        {
            "a": pa.array([1, 2, 3], type=pa.int32()),
            "b": pa.array([5, 6, 7], type=pa.int32()),
        },
    )

    @asset
    def create_table(delta_table: DeltaTableResource):
        write_deltalake(
            delta_table.url,
            data,
            storage_options=delta_table.storage_options.model_dump(),
        )
        write_deltalake(
            delta_table.url,
            data,
            storage_options=delta_table.storage_options.model_dump(),
            mode="append",
        )

    @asset
    def read_table(delta_table: DeltaTableResource):
        res = delta_table.load().to_pyarrow_table()
        assert res.equals(data)

    materialize(
        [create_table, read_table],
        resources={
            "delta_table": DeltaTableResource(
                url=os.path.join(tmp_path, "table"),
                storage_options=LocalConfig(),
                client_options=ClientConfig(
                    max_retries=10,
                    retry_timeout="10s",
                    backoff_config=BackoffConfig(init_backoff="10s", base=1.2),
                ),
                version=0,
            ),
        },
    )



from collections.abc import Sequence
from typing import Union

import dagster as dg
from arro3.core import RecordBatchReader, Table
from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
from dagster._core.storage.db_io_manager import DbTypeHandler

from dagster_delta._handler.base import ArrowTypes, DeltalakeBaseArrowTypeHandler
from dagster_delta.io_manager.base import (
    BaseDeltaLakeIOManager as BaseDeltaLakeIOManager,
)


class DeltaLakePyarrowIOManager(BaseDeltaLakeIOManager):  # noqa: D101
    @staticmethod
    def type_handlers() -> Sequence[DbTypeHandler]:  # noqa: D102
        return [_DeltaLakePyArrowTypeHandler()]


class _DeltaLakePyArrowTypeHandler(DeltalakeBaseArrowTypeHandler[ArrowTypes]):  # noqa: D101
    def from_arrow(
        self,
        obj: Union[ArrowStreamExportable, ArrowArrayExportable],
        target_type: type[ArrowTypes],  # type: ignore
    ) -> ArrowTypes:  # noqa: D102 # type: ignore
        data = RecordBatchReader.from_arrow(obj)

        if target_type == Table:
            return data.read_all()

        try:
            import pyarrow as pa

            if target_type == pa.Table:
                return pa.table(data)
            elif target_type == pa.RecordBatchReader:
                return pa.RecordBatchReader.from_stream(data)
        except ImportError:
            pass
        return data

    def to_arrow(
        self,
        obj: Union[ArrowStreamExportable, ArrowArrayExportable],
    ) -> ArrowStreamExportable:  # noqa: D102
        return RecordBatchReader.from_arrow(obj)

    def get_output_stats(self, obj: ArrowTypes) -> dict[str, dg.MetadataValue]:  # noqa: ARG002 # type: ignore
        """Returns output stats to be attached to the the context.

        Args:
            obj (ArrowTypes): Union[pa.Table, pa.RecordBatchReader, ds.Dataset]

        Returns:
            Mapping[str, MetadataValue]: metadata stats
        """
        return {}

    @property
    def supported_types(self) -> Sequence[type[object]]:
        """Returns the supported dtypes for this typeHandler"""
        supported_types = [Table, RecordBatchReader]
        try:
            import pyarrow as pa

            supported_types.extend([pa.Table, pa.RecordBatchReader])
        except ImportError:
            pass

        return supported_types



import logging
from typing import Any, Optional, TypeVar, Union

from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable
from deltalake import CommitProperties, DeltaTable, WriterProperties
from deltalake.table import FilterLiteralType, TableMerger

from dagster_delta._handler.utils import create_predicate
from dagster_delta.config import MergeConfig, MergeOperationsConfig, MergeType

T = TypeVar("T")


def merge_execute(
    dt: DeltaTable,
    data: Union[ArrowStreamExportable, ArrowArrayExportable],
    merge_config: MergeConfig,
    writer_properties: Optional[WriterProperties],
    commit_properties: Optional[CommitProperties],
    merge_predicate_from_metadata: Optional[str],
    merge_operations_config: Optional[MergeOperationsConfig],
    partition_filters: Optional[list[FilterLiteralType]] = None,
) -> dict[str, Any]:
    merge_type = merge_config.merge_type
    error_on_type_mismatch = merge_config.error_on_type_mismatch

    if merge_predicate_from_metadata is not None:
        predicate = merge_predicate_from_metadata
    elif merge_config.predicate is not None:
        predicate = merge_config.predicate
    else:
        raise Exception("merge predicate was not provided")

    target_alias = merge_config.target_alias

    if partition_filters is not None:
        partition_predicate = create_predicate(partition_filters, target_alias=target_alias)

        predicate = f"{predicate} AND {partition_predicate}"
        logger = logging.getLogger()
        logger.setLevel("DEBUG")
        logger.debug("Using explicit MERGE partition predicate: \n%s", predicate)

    merger = dt.merge(
        source=data,
        predicate=predicate,
        source_alias=merge_config.source_alias,
        target_alias=target_alias,
        error_on_type_mismatch=error_on_type_mismatch,
        writer_properties=writer_properties,
        commit_properties=commit_properties,
    )

    if merge_type == MergeType.update_only:
        return merger.when_matched_update_all().execute()
    elif merge_type == MergeType.deduplicate_insert:
        return merger.when_not_matched_insert_all().execute()
    elif merge_type == MergeType.upsert:
        return merger.when_matched_update_all().when_not_matched_insert_all().execute()
    elif merge_type == MergeType.replace_delete_unmatched:
        return merger.when_matched_update_all().when_not_matched_by_source_delete().execute()
    elif merge_type == MergeType.custom:
        if merge_operations_config is not None:
            operations_config = merge_operations_config
        elif merge_config.merge_operations_config is not None:
            operations_config = merge_config.merge_operations_config
        else:
            raise Exception("merge operations config was not provided")
        operations_config = MergeOperationsConfig.model_validate(operations_config)
        return apply_merge_operations(merger, operations_config).execute()
    else:
        raise NotImplementedError


def apply_merge_operations(
    merger: TableMerger,
    operations_config: MergeOperationsConfig,
) -> TableMerger:
    if operations_config.when_not_matched_insert is not None:
        for match in operations_config.when_not_matched_insert:
            merger = merger.when_not_matched_insert(
                predicate=match.predicate,
                updates=match.updates,
            )
    if operations_config.when_not_matched_insert_all is not None:
        for match in operations_config.when_not_matched_insert_all:
            merger = merger.when_not_matched_insert_all(
                predicate=match.predicate,
                except_cols=match.except_cols,
            )
    if operations_config.when_matched_update is not None:
        for match in operations_config.when_matched_update:
            merger = merger.when_matched_update(
                predicate=match.predicate,
                updates=match.updates,
            )

    if operations_config.when_matched_update_all is not None:
        for match in operations_config.when_matched_update_all:
            merger = merger.when_matched_update_all(
                predicate=match.predicate,
                except_cols=match.except_cols,
            )

    if operations_config.when_matched_delete is not None:
        for match in operations_config.when_matched_delete:
            merger = merger.when_matched_delete(
                predicate=match.predicate,
            )

    if operations_config.when_not_matched_by_source_delete is not None:
        for match in operations_config.when_not_matched_by_source_delete:
            merger = merger.when_not_matched_by_source_delete(
                predicate=match.predicate,
            )

    if operations_config.when_not_matched_by_source_update is not None:
        for match in operations_config.when_not_matched_by_source_update:
            merger = merger.when_not_matched_by_source_update(
                updates=match.updates,
                predicate=match.predicate,
            )

    return merger



from dagster_delta.io_manager.arrow import DeltaLakePyarrowIOManager
from dagster_delta.io_manager.base import (
    BaseDeltaLakeIOManager,
    SchemaMode,
    WriteMode,
)

__all__ = [
    "WriteMode",
    "SchemaMode",
    "BaseDeltaLakeIOManager",
    "DeltaLakePyarrowIOManager",
]


try:
    from dagster_delta.io_manager.polars import DeltaLakePolarsIOManager  # noqa

    __all__.extend(["DeltaLakePolarsIOManager"])

except ImportError as e:
    if "polars" in str(e):
        pass
    else:
        raise e



import os

import pyarrow as pa
import pytest
from dagster import (
    Out,
    graph,
    op,
)
from deltalake import DeltaTable

from dagster_delta import DeltaLakePyarrowIOManager, LocalConfig


@pytest.fixture
def io_manager(tmp_path) -> DeltaLakePyarrowIOManager:
    return DeltaLakePyarrowIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
    )


@op(
    out=Out(
        metadata={
            "schema": "a_df",
            "mode": "append",
            "commit_properties": {"custom_metadata": {"userName": "John Doe"}},
        },
    ),
)
def a_df() -> pa.Table:
    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@graph
def to_one_df():
    a_df()


def test_deltalake_io_manager_with_ops_rust_writer(tmp_path, io_manager):
    resource_defs = {"io_manager": io_manager}

    job = to_one_df.to_job(resource_defs=resource_defs)

    result = [1, 2, 3]
    for _ in range(1, 4):
        res = job.execute_in_process()

        assert res.success

        dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))
        last_action = dt.history(1)[0]
        assert last_action["userName"] == "John Doe"
        out_df = dt.to_pyarrow_table()
        assert out_df["a"].to_pylist() == result

        result.extend([1, 2, 3])


@pytest.fixture
def io_manager_with_writer_metadata(tmp_path) -> DeltaLakePyarrowIOManager:
    return DeltaLakePyarrowIOManager(
        root_uri=str(tmp_path),
        storage_options=LocalConfig(),
        commit_properties={"custom_metadata": {"userName": "John Doe"}},
        writer_properties={"compression": "ZSTD"},
    )


@op(out=Out(metadata={"schema": "a_df"}))
def a_df2() -> pa.Table:
    return pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})


@graph
def to_one_df2():
    a_df2()


def test_deltalake_io_manager_with_additional_configs(tmp_path, io_manager_with_writer_metadata):
    resource_defs = {"io_manager": io_manager_with_writer_metadata}

    job = to_one_df2.to_job(resource_defs=resource_defs)
    res = job.execute_in_process()

    assert res.success

    dt = DeltaTable(os.path.join(tmp_path, "a_df/result"))

    last_action = dt.history(1)[0]
    assert last_action["userName"] == "John Doe"

    file = dt.get_add_actions()["path"].to_pylist()[0]
    assert os.path.splitext(os.path.splitext(file)[0])[1] == ".zstd"

    out_df = dt.to_pyarrow_table()
    assert out_df["a"].to_pylist() == [1, 2, 3]



from __future__ import annotations

import datetime as dt
from typing import List, Mapping, Sequence, Union, cast  # noqa

import pendulum
from dagster import (
    AssetKey,
    MultiPartitionKey,
    MultiPartitionsDefinition,
    PartitionsDefinition,
    TimeWindowPartitionsDefinition,
)

try:
    from dagster._core.definitions.partitions.utils import TimeWindow
except ModuleNotFoundError:
    from dagster._core.definitions.time_window_partitions import TimeWindow

from dagster._core.storage.db_io_manager import TablePartitionDimension
from pendulum import instance as pdi


def generate_multi_partitions_dimension(
    asset_partition_keys: Sequence[str],
    asset_partitions_def: MultiPartitionsDefinition | PartitionsDefinition,
    partition_expr: Mapping[str, str] | TimeWindow | None,
    asset_key: AssetKey,
) -> list[TablePartitionDimension]:
    """Generates multi partition dimensions."""
    partition_dimensions: list[TablePartitionDimension] = []
    multi_partition_key_mappings = [
        cast(MultiPartitionKey, partition_key).keys_by_dimension
        for partition_key in asset_partition_keys
    ]
    for part in asset_partitions_def.partitions_defs:  # type: ignore[attr-defined]
        partitions: list[TimeWindow | str] = []
        for multi_partition_key_mapping in multi_partition_key_mappings:
            partition_key = multi_partition_key_mapping[part.name]
            if isinstance(part.partitions_def, TimeWindowPartitionsDefinition):
                partitions.append(
                    part.partitions_def.time_window_for_partition_key(partition_key),
                )
            else:
                partitions.append(partition_key)

        partition_expr_str = partition_expr.get(part.name)  # type: ignore[attr-defined]
        if partition_expr_str is None:
            raise ValueError(
                f"Asset '{asset_key}' has partition {part.name}, but the"
                f" 'partition_expr' metadata does not contain a {part.name} entry,"
                " so we don't know what column to filter it on. Specify which"
                " column of the database contains data for the"
                f" {part.name} partition.",
            )
        partitions_: TimeWindow | Sequence[str]
        if all(isinstance(partition, TimeWindow) for partition in partitions):
            checker = MultiTimePartitionsChecker(
                partitions=cast(list[TimeWindow], partitions),
            )
            if not checker.is_consecutive():
                raise ValueError("Dates are not consecutive.")
            partitions_ = TimeWindow(
                start=checker.start,
                end=checker.end,
            )
        elif all(isinstance(partition, str) for partition in partitions):
            partitions_ = list(set(cast(list[str], partitions)))
        else:
            raise ValueError("Unknown partition type")
        partition_dimensions.append(
            TablePartitionDimension(
                partition_expr=cast(str, partition_expr_str),
                partitions=partitions_,
            ),
        )
    return partition_dimensions


def generate_single_partition_dimension(
    partition_expr: str,
    asset_partition_keys: Sequence[str],
    asset_partitions_time_window: TimeWindow | None,
) -> TablePartitionDimension:
    """Given a single partition, generate a TablePartitionDimension object that can be used to create a TableSlice object.

    Args:
        partition_expr (str): Partition expression for the asset partition
        asset_partition_keys (Sequence[str]): Partition keys for the asset
        asset_partitions_time_window (TimeWindow | None): TimeWindow object for the asset partition

    Returns:
        TablePartitionDimension: TablePartitionDimension object
    """
    partition_dimension: TablePartitionDimension
    if isinstance(asset_partitions_time_window, TimeWindow):
        partition_dimension = TablePartitionDimension(
            partition_expr=partition_expr,
            partitions=(asset_partitions_time_window if asset_partition_keys else []),  # type: ignore
        )
    else:
        partition_dimension = TablePartitionDimension(
            partition_expr=partition_expr,
            partitions=asset_partition_keys,
        )
    return partition_dimension


class MultiTimePartitionsChecker:
    def __init__(self, partitions: list[TimeWindow]):
        """Helper class that defines checks on a list of TimeWindow objects
        most importantly, partitions should be consecutive.

        Args:
            partitions (list[TimeWindow]): List of TimeWindow objects
        """
        self._partitions = partitions

        start_date = min([w.start for w in self._partitions])
        end_date = max([w.end for w in self._partitions])

        if not isinstance(start_date, dt.datetime):
            raise ValueError("Start date is not a datetime")
        if not isinstance(end_date, dt.datetime):
            raise ValueError("End date is not a datetime")

        self.start = start_date
        self.end = end_date

    @property
    def hourly_delta(self) -> int:
        deltas = [date_diff(w.start, w.end).in_hours() for w in self._partitions]
        if len(set(deltas)) != 1:
            raise ValueError(
                "TimeWindowPartitionsDefinition must have the same delta from start to end",
            )
        return int(deltas[0])

    def is_consecutive(self) -> bool:
        """Checks whether the provided start dates of each partition timewindow is consecutive"""
        expected_starts = {
            pdi(self.start).add(hours=self.hourly_delta * i)
            for i in range(len(set(self._partitions)))
        }

        actual_starts = {pdi(d.start) for d in self._partitions}

        return expected_starts == actual_starts


def date_diff(start: dt.datetime, end: dt.datetime) -> pendulum.Interval:
    """Compute an interval between two dates"""
    start_ = pendulum.instance(start)
    end_ = pendulum.instance(end)
    return end_ - start_



from collections.abc import Iterable, Sequence
from typing import Optional, Union, cast

try:
    from dagster._core.definitions.partitions.utils import (
        TimeWindow,
    )
except ImportError:
    from dagster._core.definitions.time_window_partitions import (
        TimeWindow,
    )

from dagster._core.storage.db_io_manager import TablePartitionDimension
from deltalake.schema import Field as DeltaField
from deltalake.schema import PrimitiveType, Schema
from deltalake.table import FilterLiteralType


def partition_dimensions_to_dnf(
    partition_dimensions: Iterable[TablePartitionDimension],
    table_schema: Schema,
    input_dnf: bool = False,  # during input we want to read a range when it's (un)-partitioned
    date_format: Optional[dict[str, str]] = None,
) -> Optional[list[FilterLiteralType]]:
    """Converts partition dimensions to dnf filters"""
    parts = []
    for partition_dimension in partition_dimensions:
        field = _field_from_schema(partition_dimension.partition_expr, table_schema)
        if field is None:
            raise ValueError(
                f"Field {partition_dimension.partition_expr} is not part of table schema.",
                "Currently only column names are supported as partition expressions",
            )
        if isinstance(field.type, PrimitiveType):
            if field.type.type in ["timestamp", "date"]:
                filter_ = _time_window_partition_dnf(
                    partition_dimension,
                    field.type.type,
                    input_dnf,
                )
                if isinstance(filter_, list):
                    parts.extend(filter_)
                else:
                    parts.append(filter_)
            elif field.type.type in ["string", "integer"]:
                field_date_format = date_format.get(field.name) if date_format is not None else None
                filter_ = _value_dnf(
                    partition_dimension,
                    field_date_format,
                    field.type.type,
                )
                if isinstance(filter_, list):
                    parts.extend(filter_)
                else:
                    parts.append(filter_)
            else:
                raise ValueError(f"Unsupported partition type {field.type.type}")
        else:
            raise ValueError(f"Unsupported partition type {field.type}")

    return parts if len(parts) > 0 else None


def _value_dnf(
    table_partition: TablePartitionDimension,
    date_format: Optional[str] = None,
    field_type: Optional[str] = None,
) -> Union[
    list[tuple[str, str, Union[int, str]]],
    tuple[str, str, Sequence[str]],
    tuple[str, str, str],
]:  # noqa: ANN202
    # ", ".join(f"'{partition}'" for partition in table_partition.partitions)  # noqa: ERA001
    if (
        isinstance(table_partition.partitions, list)
        and all(isinstance(p, TimeWindow) for p in table_partition.partitions)
    ) or isinstance(table_partition.partitions, TimeWindow):
        if date_format is None:
            raise Exception(
                "Date format not set on time based partition definition, even though field is (str, int). Set date fmt on the partition_def, or change column type to date/datetime.",
            )
        if isinstance(table_partition.partitions, list):
            start_dts = [partition.start for partition in table_partition.partitions]  # type: ignore
            end_dts = [partition.end for partition in table_partition.partitions]  # type: ignore
            start_dt = min(start_dts)
            end_dt = max(end_dts)
        else:
            start_dt = table_partition.partitions.start  # type: ignore[attr-defined]
            end_dt = table_partition.partitions.end  # type: ignore[attr-defined]

        start_dt = start_dt.strftime(date_format)
        end_dt = end_dt.strftime(date_format)

        if field_type == "integer":
            start_dt = int(start_dt)
            end_dt = int(end_dt)
        return [
            (table_partition.partition_expr, ">=", start_dt),
            (table_partition.partition_expr, "<", end_dt),
        ]

    else:
        partition = cast(Sequence[str], table_partition.partitions)
        partition = list(set(partition))
        if len(partition) > 1:
            return (table_partition.partition_expr, "in", partition)
        else:
            return (table_partition.partition_expr, "=", partition[0])


def _time_window_partition_dnf(
    table_partition: TablePartitionDimension,
    data_type: str,
    input_dnf: bool,
) -> Union[FilterLiteralType, list[FilterLiteralType]]:
    if isinstance(table_partition.partitions, list):
        raise Exception(
            "For date primitive we shouldn't have received a sequence[str] but a TimeWindow",
        )
    else:
        partition = cast(TimeWindow, table_partition.partitions)
        start_dt, end_dt = partition
        start_dt, end_dt = start_dt.replace(tzinfo=None), end_dt.replace(tzinfo=None)

    if data_type == "date":
        start_dt, end_dt = (
            start_dt.date(),
            end_dt.date(),
        )

    if input_dnf:
        return [
            (table_partition.partition_expr, ">=", start_dt),
            (table_partition.partition_expr, "<", end_dt),
        ]
    else:
        return (table_partition.partition_expr, "=", start_dt)


def _field_from_schema(field_name: str, schema: Schema) -> Optional[DeltaField]:
    for field in schema.fields:
        if field.name == field_name:
            return field
    return None



import datetime as dt

import pytest
from dagster import AssetKey, MultiPartitionKey, MultiPartitionsDefinition, TimeWindow
from dagster._core.storage.db_io_manager import TablePartitionDimension

from dagster_delta._db_io_manager import utils


# NB: dagster uses dt.datetime even for dates
@pytest.fixture
def daily_partitions_time_window_consecutive() -> list[TimeWindow]:
    return [
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 0),
            end=dt.datetime(2022, 1, 2, 0),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 2, 0),
            end=dt.datetime(2022, 1, 3, 0),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 3, 0),
            end=dt.datetime(2022, 1, 4, 0),
        ),
    ]


@pytest.fixture
def daily_partitions_time_window_not_consecutive() -> list[TimeWindow]:
    return [
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 0),
            end=dt.datetime(2022, 1, 2, 0),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 2, 0),
            end=dt.datetime(2022, 1, 3, 0),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 4, 0),
            end=dt.datetime(2022, 1, 5, 0),
        ),
    ]


@pytest.fixture
def hourly_partitions_time_window_consecutive() -> list[TimeWindow]:
    return [
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 1),
            end=dt.datetime(2022, 1, 1, 2),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 2),
            end=dt.datetime(2022, 1, 1, 3),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 3),
            end=dt.datetime(2022, 1, 1, 4),
        ),
    ]


@pytest.fixture
def hourly_partitions_time_window_not_consecutive() -> list[TimeWindow]:
    return [
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 1),
            end=dt.datetime(2022, 1, 1, 2),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 2),
            end=dt.datetime(2022, 1, 1, 3),
        ),
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 4),
            end=dt.datetime(2022, 1, 1, 5),
        ),
    ]


# NB: The MonthlyPartitionDefinition in dagster will orchestrate the partitions into separate jobs to avoid issues with different month duration in hours so we only need to check a single month at a time. We need to check a single month doesn't collide witht the code like in the case of PR #20
@pytest.fixture
def monthly_partitions_time_window() -> list[TimeWindow]:
    return [
        TimeWindow(
            start=dt.datetime(2022, 1, 1, 0),
            end=dt.datetime(2022, 2, 1, 0),
        ),
    ]


def test_multi_time_partitions_monthly_checker(
    monthly_partitions_time_window: list[TimeWindow],
):
    checker = utils.MultiTimePartitionsChecker(
        partitions=monthly_partitions_time_window,
    )

    assert checker.hourly_delta == 744
    assert checker.start == dt.datetime(2022, 1, 1, 0)
    assert checker.end == dt.datetime(2022, 2, 1, 0)
    assert checker.is_consecutive()


def test_multi_time_partitions_daily_checker_consecutive(
    daily_partitions_time_window_consecutive: list[TimeWindow],
):
    checker = utils.MultiTimePartitionsChecker(
        partitions=daily_partitions_time_window_consecutive,
    )

    assert checker.hourly_delta == 24
    assert checker.start == dt.datetime(2022, 1, 1, 0)
    assert checker.end == dt.datetime(2022, 1, 4, 0)
    assert checker.is_consecutive()


def test_multi_time_partitions_daily_checker_non_consecutive(
    daily_partitions_time_window_not_consecutive: list[TimeWindow],
):
    checker = utils.MultiTimePartitionsChecker(
        partitions=daily_partitions_time_window_not_consecutive,
    )

    assert checker.hourly_delta == 24
    assert checker.start == dt.datetime(2022, 1, 1, 0)
    assert checker.end == dt.datetime(2022, 1, 5, 0)
    assert not checker.is_consecutive()


def test_multi_time_partitions_hourly_checker_consecutive(
    hourly_partitions_time_window_consecutive: list[TimeWindow],
):
    checker = utils.MultiTimePartitionsChecker(
        partitions=hourly_partitions_time_window_consecutive,
    )

    assert checker.hourly_delta == 1
    assert checker.start == dt.datetime(2022, 1, 1, 1)
    assert checker.end == dt.datetime(2022, 1, 1, 4)
    assert checker.is_consecutive()


def test_multi_time_partitions_hourly_checker_non_consecutive(
    hourly_partitions_time_window_not_consecutive: list[TimeWindow],
):
    checker = utils.MultiTimePartitionsChecker(
        partitions=hourly_partitions_time_window_not_consecutive,
    )

    assert checker.hourly_delta == 1
    assert checker.start == dt.datetime(2022, 1, 1, 1)
    assert checker.end == dt.datetime(2022, 1, 1, 5)
    assert not checker.is_consecutive()


def test_generate_single_partition_dimension_static():
    partition_dimension = utils.generate_single_partition_dimension(
        partition_expr="color_column",
        asset_partition_keys=["red"],
        asset_partitions_time_window=None,
    )
    assert isinstance(partition_dimension, TablePartitionDimension)
    assert partition_dimension.partition_expr == "color_column"
    assert partition_dimension.partitions == ["red"]


def test_generate_single_partition_dimension_time_window():
    partition_dimension = utils.generate_single_partition_dimension(
        partition_expr="date_column",
        asset_partition_keys=["2022-01-01"],
        asset_partitions_time_window=TimeWindow(
            start=dt.datetime(2022, 1, 1, 0),
            end=dt.datetime(2022, 1, 2, 0),
        ),
    )
    assert isinstance(partition_dimension, TablePartitionDimension)
    assert isinstance(partition_dimension.partitions, TimeWindow)
    assert partition_dimension.partition_expr == "date_column"
    assert partition_dimension.partitions.start == dt.datetime(2022, 1, 1, 0)
    assert partition_dimension.partitions.end == dt.datetime(2022, 1, 2, 0)


def test_generate_partition_dimensions_color_varying(
    multi_partition_with_color: MultiPartitionsDefinition,
):
    partition_dimensions = utils.generate_multi_partitions_dimension(
        asset_key=AssetKey("my_asset"),
        # NB: these must be multi partition keys
        asset_partition_keys=[
            MultiPartitionKey(keys_by_dimension={"color": "red", "date": "2022-01-01"}),
            MultiPartitionKey(
                keys_by_dimension={"color": "blue", "date": "2022-01-01"},
            ),
            MultiPartitionKey(
                keys_by_dimension={"color": "yellow", "date": "2022-01-01"},
            ),
        ],
        asset_partitions_def=multi_partition_with_color,
        partition_expr={
            "date": "date_column",
            "color": "color_column",
        },
    )
    assert len(partition_dimensions) == 2
    assert partition_dimensions[0].partition_expr == "color_column"
    assert partition_dimensions[1].partition_expr == "date_column"
    assert partition_dimensions[1].partitions.start == dt.datetime(  # type: ignore
        2022,
        1,
        1,
        0,
        tzinfo=dt.timezone.utc,
    )
    assert partition_dimensions[1].partitions.end == dt.datetime(  # type: ignore
        2022,
        1,
        2,
        0,
        tzinfo=dt.timezone.utc,
    )
    assert sorted(partition_dimensions[0].partitions) == ["blue", "red", "yellow"]


def test_generate_partition_dimensions_date_varying(
    multi_partition_with_color: MultiPartitionsDefinition,
):
    partition_dimensions = utils.generate_multi_partitions_dimension(
        asset_key=AssetKey("my_asset"),
        # NB: these must be multi partition keys
        asset_partition_keys=[
            MultiPartitionKey(keys_by_dimension={"color": "red", "date": "2022-01-01"}),
            MultiPartitionKey(keys_by_dimension={"color": "red", "date": "2022-01-02"}),
            MultiPartitionKey(keys_by_dimension={"color": "red", "date": "2022-01-03"}),
        ],
        asset_partitions_def=multi_partition_with_color,
        partition_expr={
            "date": "date_column",
            "color": "color_column",
        },
    )
    assert len(partition_dimensions) == 2
    assert partition_dimensions[0].partition_expr == "color_column"
    assert partition_dimensions[1].partition_expr == "date_column"
    assert partition_dimensions[1].partitions.start == dt.datetime(  # type: ignore
        2022,
        1,
        1,
        0,
        tzinfo=dt.timezone.utc,
    )
    assert partition_dimensions[1].partitions.end == dt.datetime(  # type: ignore
        2022,
        1,
        4,
        0,
        tzinfo=dt.timezone.utc,
    )
    assert partition_dimensions[0].partitions == ["red"]
